{#####################################################################################}
{##                                                                                 ##}
{## ABC - Advanced Blocksorting Compressor                                          ##}
{##                                                                                 ##}
{## ABC_AWFC_Unit                                                                   ##}
{##                                                                                 ##}
{## Advanced Weighted Frequency Count transformation (MTF replacement)              ##}
{##                                                                                 ##}
{## Copyright (C) 2002-2003 Jrgen Abel                                             ##}
{##                                                                                 ##}
{#####################################################################################}


{$I ABC_Include.pas}


Unit ABC_AWFC_Unit;


Interface


Uses
  Classes, ABC_Global_Unit;


Const
  ABC_AWFC_NUMBER_OF_LEVELS       = 12;   { Number of levels }
  ABC_AWFC_VALUE_PARAM_0          = 2600; { Parameter 0 }
  ABC_AWFC_VALUE_PARAM_1          = 4185; { Parameter 1 }


Type
  T_ABC_AWFC_A_Level_Integer      = Array [0 .. ABC_AWFC_NUMBER_OF_LEVELS] of Integer;


Const
  ABC_AWFC_A_LEVEL_START          : T_ABC_AWFC_A_Level_Integer  = (0, 1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048); { Start der Stufen }
  ABC_AWFC_A_LEVEL_VALUE          : T_ABC_AWFC_A_Level_Integer  = (131072, 16384, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);       { Wert innerhalb der Stufen }


Type
  T_ABC_AWFC = Class
  Protected
    M_A_Symbol_Index          : Array [0 .. ABC_BYTE_MAXIMUM_SYMBOL] of Byte;
    M_A_List                  : Array [0 .. ABC_BYTE_MAXIMUM_SYMBOL] of Integer;
    M_A_Symbol_Value          : Array [0 .. ABC_BYTE_MAXIMUM_SYMBOL] of Integer;
    M_More_Frequent_Symbol_R  : Integer;
    M_A_Level_Value           : Array [0 .. ABC_AWFC_NUMBER_OF_LEVELS] of Integer;
    M_A_Frequency             : Array [$00 .. ABC_BYTE_MAXIMUM_SYMBOL] of Integer;
    M_Alphabet_Size           : Integer;
    M_Alphabet_Max            : Integer;
    Procedure   Calculate_Symbol_Distribution (Var Compress_Information : T_R_ABC_Compress_Information);
    Procedure   Move (P_Buffer : T_P_ABC_Byte_Array; Index : Integer);
  Public
    Procedure   Encode (Var Compress_Information : T_R_ABC_Compress_Information);
    Procedure   Decode (Var Compress_Information : T_R_ABC_Compress_Information);
  End; { T_ABC_AWFC }


Implementation


Uses
  ABC_MCP_Unit;


{-------------------------------------------------------------------------------------}
Procedure T_ABC_AWFC.Calculate_Symbol_Distribution (Var Compress_Information : T_R_ABC_Compress_Information);
{ Calculates alphabet size and share of frequent symbols                              }
{-------------------------------------------------------------------------------------}
  Var
    Index                   : Integer;
    Symbol                  : Byte;
    Average_Frequency       : Integer;
    More_Frequent_Symbol_N  : Integer;

  Begin { T_ABC_AWFC.Calculate_Symbol_Distribution }
    { Initialize array }
    FillChar (M_A_Frequency [0], SizeOf (M_A_Frequency), $00);

    { Calculate frequency distribution }
    For Index := 0 to Compress_Information.P_Input_Record^.Size - 1 do
      Begin { For }
        Symbol := Compress_Information.P_Input_Record^.P_Buffer^ [Index];
        Inc (M_A_Frequency [Symbol]);
      End; { For }

    { Calculate greatest alphabet symbol }
    M_Alphabet_Max := ABC_BYTE_MAXIMUM_SYMBOL;
    While (M_Alphabet_Max > 0) and (M_A_Frequency [M_Alphabet_Max] = 0) do
      Begin { While }
        Dec (M_Alphabet_Max);
      End; { While }
    Inc (M_Alphabet_Max);

    { Calculate alphabet size }
    M_Alphabet_Size := 0;
    For Symbol := 0 to M_Alphabet_Max - 1 do
      Begin { For }
        If M_A_Frequency [Symbol] > 0 then
          Begin { then }
            Inc (M_Alphabet_Size);
          End; { then }
      End; { For }

    { Calculate share of frequent symbols }
    Average_Frequency       := Compress_Information.P_Input_Record^.Size div M_Alphabet_Size;
    More_Frequent_Symbol_N  := 0;
    For Symbol := 0 to M_Alphabet_Max - 1 do
      Begin { For }
        If M_A_Frequency [Symbol] > (Average_Frequency shl 1) then
          Begin { then }
            Inc (More_Frequent_Symbol_N);
          End; { then }
      End; { For }
    M_More_Frequent_Symbol_R := More_Frequent_Symbol_N * 100 div M_Alphabet_Size;
  End; { T_ABC_AWFC.Calculate_Symbol_Distribution }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_AWFC.Move (P_Buffer : T_P_ABC_Byte_Array; Index : Integer);
{ Moves a symbol inside the table by recalculating the table                          }
{-------------------------------------------------------------------------------------}
  Var
    I_0             : Integer;
    I_1             : Integer;
    Index_Old       : Integer;
    Index_New       : Integer;
    Symbol_Old      : Byte;
    Symbol_New      : Byte;
    Value           : Integer;
    List_Index_Old  : Integer;
    List_Index_New  : Integer;

  Begin { T_ABC_AWFC.Move }
    { Book all changes of the list }
    For I_0 := 0 to ABC_AWFC_NUMBER_OF_LEVELS - 1 do
      Begin { For }
        Index_New   := Index - ABC_AWFC_A_LEVEL_START [I_0];
        Index_Old   := Index - ABC_AWFC_A_LEVEL_START [I_0 + 1];
        Value       := M_A_Level_Value [I_0];

        { Book off old value }
        If (Index_Old >= 0)  then
          Begin { then }
            Symbol_Old := P_Buffer^ [Index_Old];
            Dec (M_A_Symbol_Value [Symbol_Old], Value);

            { Insert new value }
            List_Index_Old := M_A_Symbol_Index [Symbol_Old];
            List_Index_New := List_Index_Old;
            While (List_Index_New < ABC_BYTE_MAXIMUM_SYMBOL) and (M_A_Symbol_Value [M_A_List [List_Index_New + 1]] > M_A_Symbol_Value [M_A_List [List_Index_Old]]) do
              Begin { While }
                Inc (List_Index_New);
              End; { While }
            If List_Index_New > List_Index_Old then
              Begin { then }
                { Shift list }
                For I_1 := List_Index_Old to List_Index_New - 1 do
                  Begin { For }
                    M_A_List [I_1]                    := M_A_List [I_1 + 1];
                    M_A_Symbol_Index [M_A_List [I_1]] := I_1;
                  End; { For }
                M_A_List [List_Index_New]     := Symbol_Old;
                M_A_Symbol_Index [Symbol_Old] := List_Index_New;
              End; { then }
          End; { then }

        { Book in new value }
        If Index_New >= 0 then
          Begin { then }
            Symbol_New  := P_Buffer^ [Index_New];
            Inc (M_A_Symbol_Value [Symbol_New], Value);

            { Insert new value }
            List_Index_Old := M_A_Symbol_Index [Symbol_New];
            List_Index_New := List_Index_Old;
            While (List_Index_New > 0) and (M_A_Symbol_Value [M_A_List [List_Index_New - 1]] < M_A_Symbol_Value [M_A_List [List_Index_Old]]) do
              Begin { While }
                Dec (List_Index_New);
              End; { While }
            If List_Index_New < List_Index_Old then
              Begin { then }
                { Shift list }
                For I_1 := List_Index_Old downto List_Index_New + 1 do
                  Begin { For }
                    M_A_List [I_1]                    := M_A_List [I_1 - 1];
                    M_A_Symbol_Index [M_A_List [I_1]] := I_1;
                  End; { For }
                M_A_List [List_Index_New]     := Symbol_New;
                M_A_Symbol_Index [Symbol_New] := List_Index_New;
              End; { then }
          End; { then }
      End; { For }
  End; { T_ABC_AWFC.Move }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_AWFC.Encode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Encodes a block                                                                     }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;

  Begin { T_ABC_AWFC.Encode }
    If Compress_Information.P_Input_Record^.Size <= 0 then
      Begin { then }
        { Switch buffer }
        ABC_MCP.Switch_Block (Compress_Information);
        Exit;
      End; { then }

    { Initialize arrays }
    For I := 0 to ABC_BYTE_MAXIMUM_SYMBOL do
      Begin { For }
        M_A_List [I]          := I;
        M_A_Symbol_Index [I]  := I;
        M_A_Symbol_Value [I]  := 0;
      End; { For }

    { Calculate symbol distribution }
    Calculate_Symbol_Distribution (Compress_Information);

    { Calculate weights }
    M_A_Level_Value [0] := ABC_AWFC_A_LEVEL_VALUE [0];
    M_A_Level_Value [1] := ABC_AWFC_A_LEVEL_VALUE [1];
    For I := 2 to ABC_AWFC_NUMBER_OF_LEVELS do
      Begin { For }
        M_A_Level_Value [I] := M_A_Level_Value [I - 1] * ABC_AWFC_VALUE_PARAM_0 div (ABC_AWFC_VALUE_PARAM_1 + (I * M_More_Frequent_Symbol_R * M_More_Frequent_Symbol_R));
      End; { For }

    { Output share of frequent symbols }
    Compress_Information.P_Administration_Block^.P_Buffer^ [Compress_Information.P_Administration_Block^.Cursor] := M_More_Frequent_Symbol_R;
    Inc (Compress_Information.P_Administration_Block^.Cursor);

    { Handle data }
    For I := 0 to Compress_Information.P_Input_Record^.Size - 1 do
      Begin { For }
        Compress_Information.P_Output_Record^.P_Buffer^ [I] := M_A_Symbol_Index [Compress_Information.P_Input_Record^.P_Buffer^ [I]];
        Move (Compress_Information.P_Input_Record^.P_Buffer, I);
      End; { For }

    { Lnge setzen }
    Compress_Information.P_Output_Record^.Size := Compress_Information.P_Input_Record^.Size;
  End; { T_ABC_AWFC.Encode }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_AWFC.Decode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Decodes a block                                                                     }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;

  Begin { T_ABC_AWFC.Decode }
    If Compress_Information.P_Input_Record^.Size <= 0 then
      Begin { then }
        { Switch buffer }
        ABC_MCP.Switch_Block (Compress_Information);
        Exit;
      End; { then }

    { Initialize arrays }
    For I := 0 to ABC_BYTE_MAXIMUM_SYMBOL do
      Begin { For }
        M_A_List [I]          := I;
        M_A_Symbol_Index [I]  := I;
        M_A_Symbol_Value [I]  := 0;
      End; { For }

    { Input share of frequent symbols }
    Dec (Compress_Information.P_Administration_Block^.Cursor);
    M_More_Frequent_Symbol_R := Compress_Information.P_Administration_Block^.P_Buffer^ [Compress_Information.P_Administration_Block^.Cursor];

    { Calcualte weights }
    M_A_Level_Value [0] := ABC_AWFC_A_LEVEL_VALUE [0];
    M_A_Level_Value [1] := ABC_AWFC_A_LEVEL_VALUE [1];
    For I := 2 to ABC_AWFC_NUMBER_OF_LEVELS do
      Begin { For }
        M_A_Level_Value [I] := M_A_Level_Value [I - 1] * ABC_AWFC_VALUE_PARAM_0 div (ABC_AWFC_VALUE_PARAM_1 + (I * M_More_Frequent_Symbol_R * M_More_Frequent_Symbol_R));
      End; { For }

    { Handle data }
    For I := 0 to Compress_Information.P_Input_Record^.Size - 1 do
      Begin { For }
        Compress_Information.P_Output_Record^.P_Buffer^ [I] := M_A_List [Compress_Information.P_Input_Record^.P_Buffer^ [I]];
        Move (Compress_Information.P_Output_Record^.P_Buffer, I);
      End; { For }

    { Set length }
    Compress_Information.P_Output_Record^.Size := Compress_Information.P_Input_Record^.Size;
  End; { T_ABC_AWFC.Decode }


End.

