{#####################################################################################}
{##                                                                                 ##}
{## ABC - Advanced Blocksorting Compressor                                          ##}
{##                                                                                 ##}
{## ABC_Coder_Unit                                                                  ##}
{##                                                                                 ##}
{## Arithmetic coder based on the arithmetic coder in CACM87 from Witten, Neal and  ##}
{##   Cleary and on the range coder from Martin                                     ##}
{##                                                                                 ##}
{## Copyright (C) 2002-2003 Jrgen Abel                                             ##}
{##                                                                                 ##}
{#####################################################################################}


{$I ABC_Include.pas}


Unit ABC_Coder_Unit;


Interface


Uses
  ABC_Global_Unit,
  ABC_Model_Unit;


Const
  ABC_RANGE_CODER_BIT_MAX              = 32;
  ABC_RANGE_CODER_BITS_PER_BYTE        = 8;
  ABC_RANGE_CODER_BYTE_MASK            = $FF;
  ABC_RANGE_CODER_TOP_VALUE            = Longword (1 shl (ABC_RANGE_CODER_BIT_MAX - 1));
  ABC_RANGE_CODER_BOTTOM_VALUE         = Longword (ABC_RANGE_CODER_TOP_VALUE shr ABC_RANGE_CODER_BITS_PER_BYTE);
  ABC_RANGE_CODER_SHIFT_BITS           = ABC_RANGE_CODER_BIT_MAX - 9;
  ABC_RANGE_CODER_EXTRA_BITS           = ((ABC_RANGE_CODER_BIT_MAX - 2) mod 8) + 1;
  ABC_RANGE_CODER_RANGE_MASK           = Longword (ABC_RANGE_CODER_BYTE_MASK shl ABC_RANGE_CODER_SHIFT_BITS);
  ABC_RANGE_CODER_BYTE_RESET           = 3;
  ABC_RANGE_CODER_EXPONENT_START       = 4;


Type
  T_ABC_Range_Coder = Class
  Protected
    M_P_Input_Buffer        : T_P_ABC_Byte_Array;       { Pointer to input buffer }
    M_P_Output_Buffer       : T_P_ABC_Byte_Array;       { Pointer to output buffer }
    M_Input_Index           : Integer;                  { Input index }
    M_Output_Index          : Integer;                  { Output index }
    M_Low                   : Longword;                 { Lower bound of interval }
    M_Range                 : Longword;                 { Range }
    M_Buffer                : Byte;                     { Buffer symbol }
    M_Bytes_To_Follow       : Integer;                  { Number of bytes, which are remaining }
  Public
    Procedure Start_Encoding (P_Output_Buffer : T_P_ABC_Byte_Array; Output_Index : Integer);
    Procedure Renormalize_Encoding;
    Procedure Encode_Symbol (Model : T_ABC_Model; Symbol : Integer);
    Function  Done_Encoding : Integer;
    Procedure Start_Decoding (P_Input_Buffer : T_P_ABC_Byte_Array; Input_Index : Integer);
    Procedure Renormalize_Decoding;
    Function  Decode_Symbol (Model : T_ABC_Model) : Integer;
    Function  Done_Decoding : Integer;
  End; { T_ABC_Range_Coder }

  T_ABC_Coder = Class
  Protected
    M_Range_Coder       : T_ABC_Range_Coder;
    M_Symbol_Model      : T_ABC_Symbol_Model;
    M_Byte_Model        : T_ABC_Symbol_Model;
    M_Exponent_Model    : T_ABC_Exponent_Model;
    M_Mantissa_Model    : T_ABC_Mantissa_Model;
    M_Mantissa_Model_1  : T_ABC_Mantissa_Model;
    M_Mantissa_Model_2  : T_ABC_Mantissa_Model;
    M_Mantissa_Model_3  : T_ABC_Mantissa_Model;
    Procedure   Encode_Byte (Symbol : Byte);
    Function    Decode_Byte : Byte;
    Procedure   Encode_Symbol (Symbol : Integer);
    Function    Decode_Symbol : Integer;
    Procedure   Encode_Exponent (Symbol : Byte);
    Function    Decode_Exponent : Integer;
    Procedure   Encode_Mantissa (Symbol : Integer);
    Function    Decode_Mantissa : Integer;
    Procedure   Encode_Mantissa_1 (Symbol : Integer);
    Function    Decode_Mantissa_1 : Integer;
    Procedure   Encode_Mantissa_2 (Symbol : Integer);
    Function    Decode_Mantissa_2 : Integer;
    Procedure   Encode_Mantissa_3 (Symbol : Integer);
    Function    Decode_Mantissa_3 : Integer;
  Public
    Constructor Create;
    Destructor  Destroy; Override;
    Procedure   Encode (Var Compress_Information : T_R_ABC_Compress_Information);
    Procedure   Decode (Var Compress_Information : T_R_ABC_Compress_Information);
  End; { T_ABC_Coder }


Implementation


Uses
  ABC_MCP_Unit,
  Math;


{#####################################################################################}
{ T_ABC_Range_Coder                                                                   }
{#####################################################################################}


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Range_Coder.Start_Encoding (P_Output_Buffer : T_P_ABC_Byte_Array; Output_Index : Integer);
{ Initializing coder                                                                  }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Range_Coder.Start_Encoding }
    M_P_Output_Buffer := P_Output_Buffer;

    M_Output_Index    := Output_Index;
    M_Low             := 0;
    M_Range           := ABC_RANGE_CODER_TOP_VALUE;
    M_Buffer          := 0;
    M_Bytes_To_Follow := 0;
  End; { T_ABC_Range_Coder.Start_Encoding }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Range_Coder.Renormalize_Encoding;
{ Renormalizing coder                                                                 }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Range_Coder.Renormalize_Encoding }
    While M_Range <= ABC_RANGE_CODER_BOTTOM_VALUE do
      Begin { While }
        If M_Low < ABC_RANGE_CODER_RANGE_MASK then
          Begin { then }
            M_P_Output_Buffer^ [M_Output_Index] := M_Buffer;
            Inc (M_Output_Index);

            While M_Bytes_To_Follow > 0 do
              Begin { While }
                M_P_Output_Buffer^ [M_Output_Index] := ABC_RANGE_CODER_BYTE_MASK;
                Inc (M_Output_Index);
                Dec (M_Bytes_To_Follow);
              End; { While }

            M_Buffer := M_Low shr ABC_RANGE_CODER_SHIFT_BITS;
          End { then }
            else
          Begin { else }
            If (M_Low and ABC_RANGE_CODER_TOP_VALUE) <> $00 then
              Begin { then }
                M_P_Output_Buffer^ [M_Output_Index] := M_Buffer + 1;
                Inc (M_Output_Index);

                While M_Bytes_To_Follow > 0 do
                  Begin { While }
                    M_P_Output_Buffer^ [M_Output_Index] := $00;
                    Inc (M_Output_Index);
                    Dec (M_Bytes_To_Follow);
                  End; { While }

                M_Buffer := M_Low shr ABC_RANGE_CODER_SHIFT_BITS;
              End { then }
                else
              Begin { else }
                Inc (M_Bytes_To_Follow);
              End; { else }
          End; { else }
        M_Range := M_Range shl ABC_RANGE_CODER_BITS_PER_BYTE;
        M_Low   := (M_Low shl ABC_RANGE_CODER_BITS_PER_BYTE) and (ABC_RANGE_CODER_TOP_VALUE - 1);
      End; { While }
  End; { T_ABC_Range_Coder.Renormalize_Encoding }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Range_Coder.Encode_Symbol (Model : T_ABC_Model; Symbol : Integer);
{ Coding of a symbol                                                                  }
{-------------------------------------------------------------------------------------}
  Var
    Index : Integer;
    Range : Longword;
    Low   : Longword;

  Begin { T_ABC_Range_Coder.Encode_Symbol }
    Index := Model.M_A_Symbol_To_Index [Symbol];

    If M_Range <= ABC_RANGE_CODER_BOTTOM_VALUE then
      Begin { then }
        Renormalize_Encoding;
      End; { then }

    Range := M_Range div Longword (Model.M_A_Cumulated_Frequency [0]);
    If Model.M_A_Cumulated_Frequency [Index - 1] < Model.M_A_Cumulated_Frequency [0] then
      Begin { then }
        Low     := Range * Longword (Model.M_A_Cumulated_Frequency [Index]);
        M_Low   := M_Low + Low;
        M_Range := Range * Longword (Model.M_A_Cumulated_Frequency [Index - 1] - Model.M_A_Cumulated_Frequency [Index]);
      End { then }
        else
      Begin { else }
        Low     := Range * Longword (Model.M_A_Cumulated_Frequency [Index]);
        M_Low   := M_Low + Low;
        M_Range := M_Range - Low;
      End; { else }

    Model.Update (Index); { Adaptives Modell anpassen }
  End; { T_ABC_Range_Coder.Encode_Symbol }


{-------------------------------------------------------------------------------------}
Function T_ABC_Range_Coder.Done_Encoding : Integer;
{ Finishing of coding and flushing of remaining bytes                                 }
{-------------------------------------------------------------------------------------}
  Var
    Temp  : Longword;

  Begin { T_ABC_Range_Coder.Done_Encoding }
    If M_Range <= ABC_RANGE_CODER_BOTTOM_VALUE then
      Begin { then }
        Renormalize_Encoding;
      End; { then }

    Temp := (M_Low shr ABC_RANGE_CODER_SHIFT_BITS) + 1;
    If Temp > ABC_RANGE_CODER_BYTE_MASK then
      Begin { then }
        M_P_Output_Buffer^ [M_Output_Index] := M_Buffer + 1;
        Inc (M_Output_Index);

        While M_Bytes_To_Follow > 0 do
          Begin { While }
            M_P_Output_Buffer^ [M_Output_Index] := $00;
            Inc (M_Output_Index);
            Dec (M_Bytes_To_Follow);
          End; { While }
      End { then }
        else
      Begin { else }
        M_P_Output_Buffer^ [M_Output_Index] := M_Buffer;
        Inc (M_Output_Index);

        While M_Bytes_To_Follow > 0 do
          Begin { While }
            M_P_Output_Buffer^ [M_Output_Index] := ABC_RANGE_CODER_BYTE_MASK;
            Inc (M_Output_Index);
            Dec (M_Bytes_To_Follow);
          End; { While }
      End; { else }

    M_P_Output_Buffer^ [M_Output_Index] := Temp and ABC_RANGE_CODER_BYTE_MASK;
    Inc (M_Output_Index);

    Result := M_Output_Index;
  End; { T_ABC_Range_Coder.Done_Encoding }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Range_Coder.Start_Decoding (P_Input_Buffer : T_P_ABC_Byte_Array; Input_Index : Integer);
{ Initializing of decoder                                                             }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Range_Coder.Start_Decoding }
    M_P_Input_Buffer  := P_Input_Buffer;
    M_Input_Index     := Input_Index + 1;

    M_Buffer          := M_P_Input_Buffer^ [M_Input_Index];
    Inc (M_Input_Index);

    M_Low             := M_Buffer shr (ABC_RANGE_CODER_BITS_PER_BYTE - ABC_RANGE_CODER_EXTRA_BITS);
    M_Range           := $01 shl ABC_RANGE_CODER_EXTRA_BITS;
  End; { T_ABC_Range_Coder.Start_Decoding }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Range_Coder.Renormalize_Decoding;
{ Renormalizing of decoder                                                            }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Range_Coder.Renormalize_Decoding }
    While M_Range <= ABC_RANGE_CODER_BOTTOM_VALUE do
      Begin { While }
        M_Low     := (M_Low shl ABC_RANGE_CODER_BITS_PER_BYTE) or ((M_Buffer shl ABC_RANGE_CODER_EXTRA_BITS) and ABC_RANGE_CODER_BYTE_MASK);

        M_Buffer  := M_P_Input_Buffer^ [M_Input_Index];
        Inc (M_Input_Index);

        M_Low     := M_Low  or (M_Buffer shr (ABC_RANGE_CODER_BITS_PER_BYTE - ABC_RANGE_CODER_EXTRA_BITS));
        M_Range   := M_Range shl ABC_RANGE_CODER_BITS_PER_BYTE;
      End; { While }
  End; { T_ABC_Range_Coder.Renormalize_Decoding }


{-------------------------------------------------------------------------------------}
Function T_ABC_Range_Coder.Decode_Symbol (Model : T_ABC_Model) : Integer;
{ Decoding of symbol                                                                  }
{-------------------------------------------------------------------------------------}
  Var
    Range                     : Longword;
    Temp                      : Longword;
    Cum                       : Longword;
    Index                     : Integer;
    D                         : Longword;
    Cumulated_Frequency_0     : Longword;

  Begin { T_ABC_Range_Coder.Decode_Symbol }
    If M_Range <= ABC_RANGE_CODER_BOTTOM_VALUE then
      Begin { then }
        Renormalize_Decoding;
      End; { then }

    Cumulated_Frequency_0 := Longword (Model.M_A_Cumulated_Frequency [0]);

    Range := M_Range div Cumulated_Frequency_0;
    Temp  := M_Low div Range;
    If Temp >= Cumulated_Frequency_0 then
      Begin { then }
        Cum := Cumulated_Frequency_0 - 1;
      End { then }
        else
      Begin { else }
        Cum := Temp;
      End; { else }

    { Calculating index for cumulated frequencies }
    Index := 1;
    While Longword (Model.M_A_Cumulated_Frequency [Index]) > Cum do
      Begin { While }
        Inc (Index);
      End; { While }

    Result  := Model.M_A_Index_To_Symbol [Index];
    Cum     := Model.M_A_Cumulated_Frequency [Index];

    Temp  := Range * Cum;
    M_Low := M_Low - Temp;
    D     := Longword (Model.M_A_Cumulated_Frequency [Index - 1]) - Cum;
    If Cum + D < Cumulated_Frequency_0 then
      Begin { then }
        M_Range := Range * D;
      End { then }
        else
      Begin { else }
        M_Range := M_Range - Temp;
      End; { else }

    Model.Update (Index); { Adaptives Modell anpassen }
  End; { T_ABC_Range_Coder.Decode_Symbol }


{-------------------------------------------------------------------------------------}
Function T_ABC_Range_Coder.Done_Decoding : Integer;
{ Finishing of decoding and flushing of remaining bytes                               }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Range_Coder.Done_Decoding }
    If M_Range <= ABC_RANGE_CODER_BOTTOM_VALUE then
      Begin { then }
        Renormalize_Decoding;
      End; { then }

    Result := M_Input_Index - ABC_RANGE_CODER_BYTE_RESET;
  End; { T_ABC_Range_Coder.Done_Decoding }




{#####################################################################################}
{ T_ABC_Coder                                                                      }
{#####################################################################################}


{-------------------------------------------------------------------------------------}
Constructor T_ABC_Coder.Create;
{ Creates all models                                                                  }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Coder.Create }
    M_Range_Coder       := T_ABC_Range_Coder.Create;
    M_Byte_Model        := T_ABC_Symbol_Model.Create (ABC_BYTE_MAXIMUM_SYMBOL + 1);
    M_Exponent_Model    := T_ABC_Exponent_Model.Create;
    M_Mantissa_Model    := T_ABC_Mantissa_Model.Create;
    M_Mantissa_Model_1  := T_ABC_Mantissa_Model_1.Create;
    M_Mantissa_Model_2  := T_ABC_Mantissa_Model_2.Create;
    M_Mantissa_Model_3  := T_ABC_Mantissa_Model_3.Create;
  End; { T_ABC_Coder.Create }

{-------------------------------------------------------------------------------------}
Destructor T_ABC_Coder.Destroy;
{ Frees all models                                                                    }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Coder.Destroy }
    M_Mantissa_Model_3.Free;
    M_Mantissa_Model_2.Free;
    M_Mantissa_Model_1.Free;
    M_Mantissa_Model.Free;
    M_Exponent_Model.Free;
    M_Byte_Model.Free;
    M_Range_Coder.Free;

    Inherited Destroy;
  End; { T_ABC_Coder.Destroy }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Coder.Encode_Byte (Symbol : Byte);
{ Codes a byte with a hierarchical model                                              }
{ 1. Determination of levels                                                          }
{ 2. Coding of 1. and 2. level                                                        }
{-------------------------------------------------------------------------------------}
  Var
    Model_0   : T_ABC_Model;
    Model_1   : T_ABC_Model;
    Symbol_0  : Integer;
    Symbol_1  : Integer;
    Model_2   : T_ABC_Model;
    Symbol_2  : Integer;

  Begin { T_ABC_Coder.Encode_Byte }
    M_Byte_Model.Get_Hierarchical_Information (Symbol, Model_0, Symbol_0, Model_1, Symbol_1, Model_2, Symbol_2);

    { Coding of 1. level }
    M_Range_Coder.Encode_Symbol (Model_0, Symbol_0);

    { Coding of 2. level }
    If Model_1 <> NIL then
      Begin { then }
        M_Range_Coder.Encode_Symbol (Model_1, Symbol_1);
      End; { then }

    { Coding of offset in 2. level }
    If Model_2 <> NIL then
      Begin { then }
        M_Range_Coder.Encode_Symbol (Model_2, Symbol_2);
      End; { then }
  End; { T_ABC_Coder.Encode_Byte }


{-------------------------------------------------------------------------------------}
Function T_ABC_Coder.Decode_Byte : Byte;
{ Decodes a byte with a hierarchical model                                            }
{ 1. Determination of levels                                                          }
{ 2. Decoding of 1. and 2. level                                                      }
{-------------------------------------------------------------------------------------}
  Var
    Second_Level_Group_Number : Integer;
    Second_Level_Group_Index  : Integer;

  Begin { T_ABC_Coder.Decode_Byte }
    Result := M_Range_Coder.Decode_Symbol (M_Byte_Model.First_Level_Model);
    If Result = ABC_MODEL_FIRST_LEVEL_ESCAPE_SYMBOL then
      Begin { then }
        Second_Level_Group_Number := M_Range_Coder.Decode_Symbol (M_Byte_Model.Second_Level_Model);
        If M_Byte_Model.A_Third_Level_Model [Second_Level_Group_Number].Number_Of_Symbols > 1 then
          Begin { then }
            Second_Level_Group_Index  := M_Range_Coder.Decode_Symbol (M_Byte_Model.A_Third_Level_Model [Second_Level_Group_Number]);
          End { then }
            else
          Begin { else }
            Second_Level_Group_Index  := 0;
          End; { else }
        Result := M_Byte_Model.Third_Level_Group_Offset_To_Symbol (Second_Level_Group_Number, Second_Level_Group_Index);
      End; { then }
  End; { T_ABC_Coder.Decode_Byte }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Coder.Encode_Symbol (Symbol : Integer);
{ Codes a symbol with a hierarchical model                                            }
{ 1. Determination of levels                                                          }
{ 2. Coding of 1. and 2. level                                                        }
{-------------------------------------------------------------------------------------}
  Var
    Model_0   : T_ABC_Model;
    Model_1   : T_ABC_Model;
    Symbol_0  : Integer;
    Symbol_1  : Integer;
    Model_2   : T_ABC_Model;
    Symbol_2  : Integer;

  Begin { T_ABC_Coder.Encode_Symbol }
    M_Symbol_Model.Get_Hierarchical_Information (Symbol, Model_0, Symbol_0, Model_1, Symbol_1, Model_2, Symbol_2);

    { Coding of 1. level }
    M_Range_Coder.Encode_Symbol (Model_0, Symbol_0);

    { Coding of 2. level }
    If Model_1 <> NIL then
      Begin { then }
        M_Range_Coder.Encode_Symbol (Model_1, Symbol_1);
      End; { then }

    { Coding of offset in 2. level }
    If Model_2 <> NIL then
      Begin { then }
        M_Range_Coder.Encode_Symbol (Model_2, Symbol_2);
      End; { then }
  End; { T_ABC_Coder.Encode_Symbol }


{-------------------------------------------------------------------------------------}
Function T_ABC_Coder.Decode_Symbol : Integer;
{ Decodes a symbol with a hierarchical model                                          }
{ 1. Determination of levels                                                          }
{ 2. Decoding of 1. and 2. level                                                      }
{-------------------------------------------------------------------------------------}
  Var
    Second_Level_Group_Number : Integer;
    Second_Level_Group_Index  : Integer;

  Begin { T_ABC_Coder.Decode_Symbol }
    Result := M_Range_Coder.Decode_Symbol (M_Symbol_Model.First_Level_Model);
    If Result = ABC_MODEL_FIRST_LEVEL_ESCAPE_SYMBOL then
      Begin { then }
        Second_Level_Group_Number := M_Range_Coder.Decode_Symbol (M_Symbol_Model.Second_Level_Model);
        If M_Symbol_Model.A_Third_Level_Model [Second_Level_Group_Number].Number_Of_Symbols > 1 then
          Begin { then }
            Second_Level_Group_Index  := M_Range_Coder.Decode_Symbol (M_Symbol_Model.A_Third_Level_Model [Second_Level_Group_Number]);
          End { then }
            else
          Begin { else }
            Second_Level_Group_Index  := 0;
          End; { else }
        Result := M_Symbol_Model.Third_Level_Group_Offset_To_Symbol (Second_Level_Group_Number, Second_Level_Group_Index);
      End; { then }
  End; { T_ABC_Coder.Decode_Symbol }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Coder.Encode_Exponent (Symbol : Byte);
{ Codes an exponent                                                                   }
{-------------------------------------------------------------------------------------}
  Var
    Model_0   : T_ABC_Model;
    Model_1   : T_ABC_Model;
    Symbol_0  : Integer;
    Symbol_1  : Integer;

  Begin { T_ABC_Coder.Encode_Exponent }
    M_Exponent_Model.Get_Hierarchical_Information (Symbol, Model_0, Symbol_0, Model_1, Symbol_1);

    { Coding of 1. level }
    M_Range_Coder.Encode_Symbol (Model_0, Symbol_0);

    { Coding of 2. level }
    If Model_1 <> NIL then
      Begin { then }
        M_Range_Coder.Encode_Symbol (Model_1, Symbol_1);
      End; { then }
  End; { T_ABC_Coder.Encode_Exponent }


{-------------------------------------------------------------------------------------}
Function T_ABC_Coder.Decode_Exponent : Integer;
{ Decodes an exponent                                                                 }
{-------------------------------------------------------------------------------------}
  Var
    Second_Level_Group_Index : Integer;

  Begin { T_ABC_Coder.Decode_Exponent }
    Result := M_Range_Coder.Decode_Symbol (M_Exponent_Model.First_Level_Model);
    If M_Exponent_Model.A_Second_Level_Model [Result].Number_Of_Symbols > 1 then
      Begin { then }
        Second_Level_Group_Index  := M_Range_Coder.Decode_Symbol (M_Exponent_Model.A_Second_Level_Model [Result]);
        Result                    := M_Exponent_Model.Second_Level_Group_Offset_To_Symbol (Result, Second_Level_Group_Index);
      End; { then }
  End; { T_ABC_Coder.Decode_Exponent }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Coder.Encode_Mantissa (Symbol : Integer);
{ Codes a mantissa bit in context "0"                                                 }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Coder.Encode_Mantissa }
    M_Range_Coder.Encode_Symbol (M_Mantissa_Model.Model, Symbol);
  End; { T_ABC_Coder.Encode_Mantissa }


{-------------------------------------------------------------------------------------}
Function T_ABC_Coder.Decode_Mantissa : Integer;
{ Decodes a mantissa bit in context "0"                                               }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Coder.Decode_Mantissa }
    Result := M_Range_Coder.Decode_Symbol (M_Mantissa_Model.Model);
  End; { T_ABC_Coder.Decode_Mantissa }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Coder.Encode_Mantissa_1 (Symbol : Integer);
{ Codes a mantissa bit in context "1"                                                 }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Coder.Encode_Mantissa_1 }
    M_Range_Coder.Encode_Symbol (M_Mantissa_Model_1.Model, Symbol);
  End; { T_ABC_Coder.Encode_Mantissa_1 }


{-------------------------------------------------------------------------------------}
Function T_ABC_Coder.Decode_Mantissa_1 : Integer;
{ Decodes a mantissa bit in context "1"                                               }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Coder.Decode_Mantissa_1 }
    Result := M_Range_Coder.Decode_Symbol (M_Mantissa_Model_1.Model);
  End; { T_ABC_Coder.Decode_Mantissa_1 }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Coder.Encode_Mantissa_2 (Symbol : Integer);
{ Codes a mantissa bit in context "2"                                                 }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Coder.Encode_Mantissa_2 }
    M_Range_Coder.Encode_Symbol (M_Mantissa_Model_2.Model, Symbol);
  End; { T_ABC_Coder.Encode_Mantissa_2 }


{-------------------------------------------------------------------------------------}
Function T_ABC_Coder.Decode_Mantissa_2 : Integer;
{ Decodes a mantissa bit in context "2"                                               }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Coder.Decode_Mantissa_2 }
    Result := M_Range_Coder.Decode_Symbol (M_Mantissa_Model_2.Model);
  End; { T_ABC_Coder.Decode_Mantissa_2 }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Coder.Encode_Mantissa_3 (Symbol : Integer);
{ Codes a mantissa bit in context "3"                                                 }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Coder.Encode_Mantissa_3 }
    M_Range_Coder.Encode_Symbol (M_Mantissa_Model_3.Model, Symbol);
  End; { T_ABC_Coder.Encode_Mantissa_3 }


{-------------------------------------------------------------------------------------}
Function T_ABC_Coder.Decode_Mantissa_3 : Integer;
{ Decodes a mantissa bit in context "3"                                               }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Coder.Decode_Mantissa_3 }
    Result := M_Range_Coder.Decode_Symbol (M_Mantissa_Model_3.Model);
  End; { T_ABC_Coder.Decode_Mantissa_3 }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Coder.Encode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Codes a block with arithmetic coding                                                }
{-------------------------------------------------------------------------------------}
  Var
    I                             : Integer;
    Exponent                      : Byte;
    Mantissa                      : Integer;
    Mantissa_Index                : Integer;
    P_Sorted_Mantissa_Buffer      : T_P_ABC_Integer_Array;
    Mantissa_C                    : Integer;
    Bit_Pos                       : Integer;
    Bit                           : Integer;
    M_EOF_Symbol                  : Integer;
    Symbol                        : Integer;
    P_Sorted_RLE_Mantissa_Buffer  : T_P_ABC_Integer_Array;
    Block_Length                  : Integer;
    RLE_Frequency_Bit_Max         : Integer;
    N                             : Integer;

  Begin { T_ABC_Coder.Encode }
    { Initialize length }
    Compress_Information.P_Output_Record^.Size := 0;

    { Check, which kind of coding to use }
    If ABC_MCP.Compress_Information.Fast_F = TRUE then
      Begin { then }
        P_Sorted_RLE_Mantissa_Buffer := Compress_Information.P_Temp_A_1;

        { Calculate RLE frequencies }
        FillChar (Compress_Information.P_Binary_Block^.A_RLE_Frequency [0], SizeOf (Compress_Information.P_Binary_Block^.A_RLE_Frequency), $00);
        For I := 0 to Compress_Information.P_Binary_Block^.RLE_Mantissa_Index - 1 do
          Begin { For }
            N         := Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [I];
            Exponent  := ABC_MCP.Get_Bin_Length (N);
            Inc (Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent]);
          End; { For }
        { Sort RLE mantissa by exponent }
        Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [0] := 0;
        For I := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
          Begin { For }
            Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I] := Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I - 1] + Compress_Information.P_Binary_Block^.A_RLE_Frequency [I - 1];
          End; { For }
        For I := 0 to Compress_Information.P_Binary_Block^.RLE_Mantissa_Index - 1 do
          Begin { For }
            { Get RLE exponent and mantissa }
            N := Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [I];
            If N > 0 then
              Begin { then }
                { Sort mantissa }
                Exponent := ABC_MCP.Get_Bin_Length (N);
                Mantissa := (N + 1) - ($01 shl Exponent);
                P_Sorted_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] := Mantissa;
                Inc (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
              End; { then }
          End; { For }
        For I := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
          Begin { For }
            Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I] := Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I - 1] + Compress_Information.P_Binary_Block^.A_RLE_Frequency [I - 1];
          End; { For }

        M_Byte_Model.Reset;
        M_Range_Coder.Start_Encoding (Compress_Information.P_Output_Record^.P_Buffer, Compress_Information.P_Output_Record^.Size);
        { Code length of exponent }
        RLE_Frequency_Bit_Max := 0;
        For I := ABC_MOST_SIGNIFICANT_BIT - 1 downto 1 do
          Begin { For }
            If Compress_Information.P_Binary_Block^.A_RLE_Frequency [I] > 0 then
              Begin { then }
                RLE_Frequency_Bit_Max := I;
                Break;
              End; { then }
          End; { For }
        Block_Length := RLE_Frequency_Bit_Max;
        Encode_Byte (Block_Length);
        { Code RLE frequencies }
        For I := 0 to RLE_Frequency_Bit_Max do
          Begin { For }
            Block_Length := Compress_Information.P_Binary_Block^.A_RLE_Frequency [I];
            Encode_Byte ((Block_Length shr $10) and $FF);
            Encode_Byte ((Block_Length shr $08) and $FF);
            Encode_Byte (Block_Length and $FF);
          End; { For }
        Compress_Information.P_Output_Record^.Size := M_Range_Coder.Done_Encoding;

        { Create models }
        M_Mantissa_Model.Reset;
        M_Mantissa_Model_1.Reset;
        M_Mantissa_Model_2.Reset;
        M_Mantissa_Model_3.Reset;
        M_Range_Coder.Start_Encoding (Compress_Information.P_Output_Record^.P_Buffer, Compress_Information.P_Output_Record^.Size);

        { Code mantissa bit 1 }
        Exponent := 1;
        For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
          Begin { For }
            Mantissa := P_Sorted_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent] + Mantissa_C];
            Encode_Mantissa_1 (Mantissa);
          End; { For }
        { Code mantissa bit 2 }
        Exponent := 2;
        For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
          Begin { For }
            Mantissa := P_Sorted_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent] + Mantissa_C];
            Encode_Mantissa_2 (Mantissa);
          End; { For }
        { Code mantissa bit 3 }
        Exponent := 3;
        For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
          Begin { For }
            Mantissa := P_Sorted_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent] + Mantissa_C];
            Encode_Mantissa_3 (Mantissa);
          End; { For }
        { Code rest of mantissa bits }
        For Exponent := ABC_RANGE_CODER_EXPONENT_START to ABC_MOST_SIGNIFICANT_BIT - 1 do
          Begin { For }
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := P_Sorted_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent] + Mantissa_C];
                Encode_Mantissa_3 (Mantissa shr (Exponent - (ABC_RANGE_CODER_EXPONENT_START - 1)));
                For Bit_Pos := Exponent - 1 downto ABC_RANGE_CODER_EXPONENT_START - 1 do
                  Begin { For }
                    Bit := (Mantissa shr (Bit_Pos - (ABC_RANGE_CODER_EXPONENT_START - 1))) and $01;
                    Encode_Mantissa (Bit);
                  End; { For }
              End; { For }
          End; { For }

        { Save length }
        Compress_Information.P_Output_Record^.Size := M_Range_Coder.Done_Encoding;

        { Code symbols }
        M_Symbol_Model  := T_ABC_Symbol_Model.Create (ABC_BYTE_MAXIMUM_SYMBOL + 1 + ABC_MODEL_EOF_RANGE);
        M_Symbol_Model.Reset;
        M_EOF_Symbol    := ABC_BYTE_MAXIMUM_SYMBOL + ABC_MODEL_EOF_RANGE;
        M_Range_Coder.Start_Encoding (Compress_Information.P_Output_Record^.P_Buffer, Compress_Information.P_Output_Record^.Size);
        I := 0;
        While I <= Compress_Information.P_Input_Record^.Size - 1 do
          Begin { While }
            Symbol := Compress_Information.P_Input_Record^.P_Buffer^ [I];
            Encode_Symbol (Symbol);

            Inc (I);
          End; { While }
        Encode_Symbol (M_EOF_Symbol);
        Compress_Information.P_Output_Record^.Size := M_Range_Coder.Done_Encoding;
        M_Symbol_Model.Free;
      End { then }
        else
      Begin { else }
        If ABC_MCP.Compress_Information.SIF_F = TRUE then
          Begin { then }
            P_Sorted_Mantissa_Buffer := Compress_Information.P_Temp_A_0;
            P_Sorted_RLE_Mantissa_Buffer := Compress_Information.P_Temp_A_1;

            { Sort mantissa data }
            Compress_Information.P_Binary_Block^.A_Mantissa_Index [0] := 0;
            For I := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                Compress_Information.P_Binary_Block^.A_Mantissa_Index [I] := Compress_Information.P_Binary_Block^.A_Mantissa_Index [I - 1] + Compress_Information.P_Binary_Block^.A_Frequency [I - 1];
              End; { For }
            Mantissa_Index := 0;
            For I := 0 to Compress_Information.P_Binary_Block^.Exponent_Index - 1 do
              Begin { For }
                { Get exponent and mantissa }
                Exponent := Compress_Information.P_Binary_Block^.P_Exponent_Buffer^ [I];
                If Exponent > 0 then
                  Begin { then }
                    { Sort mantissa }
                    Mantissa                                                                                  := Compress_Information.P_Binary_Block^.P_Mantissa_Buffer^ [Mantissa_Index];
                    P_Sorted_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent]] := Mantissa;
                    Inc (Mantissa_Index);
                    Inc (Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent]);
                  End; { then }
              End; { For }
            For I := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                Compress_Information.P_Binary_Block^.A_Mantissa_Index [I] := Compress_Information.P_Binary_Block^.A_Mantissa_Index [I - 1] + Compress_Information.P_Binary_Block^.A_Frequency [I - 1];
              End; { For }

            { Calculate RLE frequencies }
            FillChar (Compress_Information.P_Binary_Block^.A_RLE_Frequency [0], SizeOf (Compress_Information.P_Binary_Block^.A_RLE_Frequency), $00);
            For I := 0 to Compress_Information.P_Binary_Block^.RLE_Mantissa_Index - 1 do
              Begin { For }
                N         := Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [I];
                Exponent  := ABC_MCP.Get_Bin_Length (N);
                Inc (Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent]);
              End; { For }
            { Sort RLE mantissa by exponent }
            Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [0] := 0;
            For I := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I] := Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I - 1] + Compress_Information.P_Binary_Block^.A_RLE_Frequency [I - 1];
              End; { For }
            For I := 0 to Compress_Information.P_Binary_Block^.RLE_Mantissa_Index - 1 do
              Begin { For }
                { Get RLE exponent and mantissa }
                N := Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [I];
                If N > 0 then
                  Begin { then }
                    { Sort mantissa }
                    Exponent := ABC_MCP.Get_Bin_Length (N);
                    Mantissa := (N + 1) - ($01 shl Exponent);
                    P_Sorted_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] := Mantissa;
                    Inc (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
                  End; { then }
              End; { For }
            For I := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I] := Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I - 1] + Compress_Information.P_Binary_Block^.A_RLE_Frequency [I - 1];
              End; { For }

            M_Byte_Model.Reset;
            M_Range_Coder.Start_Encoding (Compress_Information.P_Output_Record^.P_Buffer, Compress_Information.P_Output_Record^.Size);
            { Code exponent index }
            Block_Length := Compress_Information.P_Binary_Block^.Exponent_Index;
            Encode_Byte ((Block_Length shr $10) and $FF);
            Encode_Byte ((Block_Length shr $08) and $FF);
            Encode_Byte (Block_Length and $FF);
            { Code bit index }
            Block_Length := Compress_Information.P_Binary_Block^.Bit_Index;
            Encode_Byte ((Block_Length shr $10) and $FF);
            Encode_Byte ((Block_Length shr $08) and $FF);
            Encode_Byte (Block_Length and $FF);
            { Code block index }
            Block_Length := Compress_Information.P_Binary_Block^.Block_Index;
            Encode_Byte ((Block_Length shr $10) and $FF);
            Encode_Byte ((Block_Length shr $08) and $FF);
            Encode_Byte (Block_Length and $FF);
            { Code length of binary blocks }
            For I := 0 to Compress_Information.P_Binary_Block^.Block_Index - 1 do
              Begin { For }
                Block_Length := Compress_Information.P_Binary_Block^.A_Block_Length [I];
                Encode_Byte ((Block_Length shr $10) and $FF);
                Encode_Byte ((Block_Length shr $08) and $FF);
                Encode_Byte (Block_Length and $FF);
              End; { For }

            { Code length of exponent }
            RLE_Frequency_Bit_Max := 0;
            For I := ABC_MOST_SIGNIFICANT_BIT - 1 downto 1 do
              Begin { For }
                If Compress_Information.P_Binary_Block^.A_RLE_Frequency [I] > 0 then
                  Begin { then }
                    RLE_Frequency_Bit_Max := I;
                    Break;
                  End; { then }
              End; { For }

            Block_Length := RLE_Frequency_Bit_Max;
            Encode_Byte (Block_Length);
            { Code RLE frequencies }
            For I := 0 to RLE_Frequency_Bit_Max do
              Begin { For }
                Block_Length := Compress_Information.P_Binary_Block^.A_RLE_Frequency [I];
                Encode_Byte ((Block_Length shr $10) and $FF);
                Encode_Byte ((Block_Length shr $08) and $FF);
                Encode_Byte (Block_Length and $FF);
              End; { For }
            Compress_Information.P_Output_Record^.Size := M_Range_Coder.Done_Encoding;

            { Code exponent }
            M_Exponent_Model.Reset;
            M_Range_Coder.Start_Encoding (Compress_Information.P_Output_Record^.P_Buffer, Compress_Information.P_Output_Record^.Size);
            For I := 0 to Compress_Information.P_Binary_Block^.Exponent_Index - 1 do
              Begin { For }
                Encode_Exponent (Compress_Information.P_Binary_Block^.P_Exponent_Buffer^ [I]);
              End; { For }
            Compress_Information.P_Output_Record^.Size := M_Range_Coder.Done_Encoding;

            { Create mantissa objects }
            M_Mantissa_Model.Reset;
            M_Mantissa_Model_1.Reset;
            M_Mantissa_Model_2.Reset;
            M_Mantissa_Model_3.Reset;
            M_Range_Coder.Start_Encoding (Compress_Information.P_Output_Record^.P_Buffer, Compress_Information.P_Output_Record^.Size);

            { Code Mantisse_1 }
            Exponent := 1;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := P_Sorted_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent] + Mantissa_C];
                Encode_Mantissa_1 (Mantissa);
              End; { For }
            { Code Mantisse_2 }
            Exponent := 2;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := P_Sorted_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent] + Mantissa_C];
                Encode_Mantissa_2 (Mantissa);
              End; { For }
            { Code Mantisse_3 }
            Exponent := 3;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := P_Sorted_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent] + Mantissa_C];
                Encode_Mantissa_3 (Mantissa);
              End; { For }
            { Code rest of mantissa bits }
            For Exponent := ABC_RANGE_CODER_EXPONENT_START to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_Frequency [Exponent] - 1 do
                  Begin { For }
                    Mantissa := P_Sorted_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent] + Mantissa_C];
                    Encode_Mantissa_3 (Mantissa shr (Exponent - (ABC_RANGE_CODER_EXPONENT_START - 1)));
                    For Bit_Pos := Exponent - 1 downto ABC_RANGE_CODER_EXPONENT_START - 1 do
                      Begin { For }
                        Bit := (Mantissa shr (Bit_Pos - (ABC_RANGE_CODER_EXPONENT_START - 1))) and $01;
                        Encode_Mantissa (Bit);
                      End; { For }
                  End; { For }
              End; { For }

            { Code RLE Mantisse_1 }
            Exponent := 1;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := P_Sorted_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent] + Mantissa_C];
                Encode_Mantissa_1 (Mantissa);
              End; { For }
            { Code RLE Mantisse_2 }
            Exponent := 2;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := P_Sorted_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent] + Mantissa_C];
                Encode_Mantissa_2 (Mantissa);
              End; { For }
            { Code RLE Mantisse_3 }
            Exponent := 3;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := P_Sorted_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent] + Mantissa_C];
                Encode_Mantissa_3 (Mantissa);
              End; { For }
            { Code rest of RLE mantissa bits }
            For Exponent := ABC_RANGE_CODER_EXPONENT_START to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
                  Begin { For }
                    Mantissa := P_Sorted_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent] + Mantissa_C];
                    Encode_Mantissa_3 (Mantissa shr (Exponent - (ABC_RANGE_CODER_EXPONENT_START - 1)));
                    For Bit_Pos := Exponent - 1 downto ABC_RANGE_CODER_EXPONENT_START - 1 do
                      Begin { For }
                        Bit := (Mantissa shr (Bit_Pos - (ABC_RANGE_CODER_EXPONENT_START - 1))) and $01;
                        Encode_Mantissa (Bit);
                      End; { For }
                  End; { For }
              End; { For }

            { Free mantissa objects }
            Compress_Information.P_Output_Record^.Size := M_Range_Coder.Done_Encoding;
          End { then }
            else
          Begin { else }
            { Code symbols }
            M_Symbol_Model := T_ABC_Symbol_Context_Model.Create (ABC_BYTE_MAXIMUM_SYMBOL + 1 + ABC_MODEL_EOF_RANGE);
            M_Symbol_Model.Reset;
            M_EOF_Symbol   := ABC_BYTE_MAXIMUM_SYMBOL + ABC_MODEL_EOF_RANGE;
            M_Range_Coder.Start_Encoding (Compress_Information.P_Output_Record^.P_Buffer, Compress_Information.P_Output_Record^.Size);
            I := 0;
            While I <= Compress_Information.P_Input_Record^.Size - 1 do
              Begin { While }
                Symbol := Compress_Information.P_Input_Record^.P_Buffer^ [I];
                Encode_Symbol (Symbol);

                Inc (I);
              End; { While }
            Encode_Symbol (M_EOF_Symbol);
            Compress_Information.P_Output_Record^.Size := M_Range_Coder.Done_Encoding;
            M_Symbol_Model.Free;
          End; { else }
      End; { else }

    { Output used block length }
    Compress_Information.P_Administration_Block^.P_Buffer^ [Compress_Information.P_Administration_Block^.Cursor] := Compress_Information.P_Output_Record^.Size and $FF;
    Inc (Compress_Information.P_Administration_Block^.Cursor);
    Compress_Information.P_Administration_Block^.P_Buffer^ [Compress_Information.P_Administration_Block^.Cursor] := (Compress_Information.P_Output_Record^.Size shr 8) and $FF;
    Inc (Compress_Information.P_Administration_Block^.Cursor);
    Compress_Information.P_Administration_Block^.P_Buffer^ [Compress_Information.P_Administration_Block^.Cursor] := (Compress_Information.P_Output_Record^.Size shr 16) and $FF;
    Inc (Compress_Information.P_Administration_Block^.Cursor);
  End; { T_ABC_Coder.Encode }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Coder.Decode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Decodes a block with arithmetic coding                                              }
{-------------------------------------------------------------------------------------}
  Var
    M_EOF_Symbol          : Integer;
    Output_Index          : Integer;
    Symbol                : Integer;
    I_0                   : Integer;
    Exponent              : Byte;
    Mantissa              : Integer;
    Mantissa_C            : Integer;
    Bit_Pos               : Integer;
    Block_Length          : Integer;
    RLE_Frequency_Bit_Max : Integer;
    I                     : Integer;

  Begin { T_ABC_Coder.Decode }
    { Read used block length }
    Dec (Compress_Information.P_Administration_Block^.Cursor);
    Compress_Information.P_Input_Record^.Size := Compress_Information.P_Administration_Block^.P_Buffer^ [Compress_Information.P_Administration_Block^.Cursor] shl 16;
    Dec (Compress_Information.P_Administration_Block^.Cursor);
    Compress_Information.P_Input_Record^.Size := Compress_Information.P_Input_Record^.Size + (Compress_Information.P_Administration_Block^.P_Buffer^ [Compress_Information.P_Administration_Block^.Cursor] shl 8);
    Dec (Compress_Information.P_Administration_Block^.Cursor);
    Compress_Information.P_Input_Record^.Size := Compress_Information.P_Input_Record^.Size + Compress_Information.P_Administration_Block^.P_Buffer^ [Compress_Information.P_Administration_Block^.Cursor];

    { Set 0 at end }
    Compress_Information.P_Input_Record^.P_Buffer^ [Compress_Information.P_Input_Record^.Size] := 0;
    Inc (Compress_Information.P_Input_Record^.Size);

    { Initialize length }
    Compress_Information.P_Input_Record^.Size := 0;

    { Check, which kind of decoding to use }
    If ABC_MCP.Compress_Information.Fast_F = TRUE then
      Begin { then }
        M_Byte_Model.Reset;
        M_Range_Coder.Start_Decoding (Compress_Information.P_Input_Record^.P_Buffer, Compress_Information.P_Input_Record^.Size);
        { Decode number of RLE mantissa bits }
        FillChar (Compress_Information.P_Binary_Block^.A_RLE_Frequency [0], SizeOf (Compress_Information.P_Binary_Block^.A_RLE_Frequency), $00);
        Block_Length          := Decode_Byte;
        RLE_Frequency_Bit_Max := Block_Length;
        { Decode mantissa bits }
        For I := 0 to RLE_Frequency_Bit_Max do
          Begin { For }
            Block_Length := Decode_Byte;
            Block_Length := (Block_Length shl $08) or Decode_Byte;
            Block_Length := (Block_Length shl $08) or Decode_Byte;
            Compress_Information.P_Binary_Block^.A_RLE_Frequency [I] := Block_Length;
          End; { For }
        Compress_Information.P_Input_Record^.Size := M_Range_Coder.Done_Decoding;

        { Calculate A_Frequency and A_Mantissa_Index }
        For I_0 := 0 to ABC_MOST_SIGNIFICANT_BIT - 1 do
          Begin { For }
            Compress_Information.P_Binary_Block^.A_Frequency [I_0]       := 0;
            Compress_Information.P_Binary_Block^.A_Mantissa_Index [I_0]  := 0;
          End; { For }
        For I_0 := 0 to Compress_Information.P_Binary_Block^.Exponent_Index - 1 do
          Begin { For }
            Exponent := Compress_Information.P_Binary_Block^.P_Exponent_Buffer^ [I_0];
            If Exponent > 0 then
              Begin { then }
                Inc (Compress_Information.P_Binary_Block^.A_Frequency [Exponent]);
              End; { then }
          End; { For }
        For I_0 := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
          Begin { For }
            Compress_Information.P_Binary_Block^.A_Mantissa_Index [I_0] := Compress_Information.P_Binary_Block^.A_Mantissa_Index [I_0 - 1] + Compress_Information.P_Binary_Block^.A_Frequency [I_0 - 1];
          End; { For }

        { Calculate A_RLE_Mantissa_Index }
        Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [0] := 0;
        For I := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
          Begin { For }
            Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I] := Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I - 1] + Compress_Information.P_Binary_Block^.A_RLE_Frequency [I - 1];
          End; { For }

        { Create mantissa objects }
        M_Mantissa_Model.Reset;
        M_Mantissa_Model_1.Reset;
        M_Mantissa_Model_2.Reset;
        M_Mantissa_Model_3.Reset;
        M_Range_Coder.Start_Decoding (Compress_Information.P_Input_Record^.P_Buffer, Compress_Information.P_Input_Record^.Size);

        { Decode Mantisse_1 }
        Exponent := 1;
        For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
          Begin { For }
            Mantissa := Decode_Mantissa_1;
            Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] := Mantissa;
            Inc (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
          End; { For }
        { Decode Mantisse_2 }
        Exponent := 2;
        For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
          Begin { For }
            Mantissa := Decode_Mantissa_2;
            Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] := Mantissa;
            Inc (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
          End; { For }
        { Decode Mantisse_3 }
        Exponent := 3;
        For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
          Begin { For }
            Mantissa := Decode_Mantissa_3;
            Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] := Mantissa;
            Inc (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
          End; { For }
        { Decode rest of mantissa bits }
        For Exponent := ABC_RANGE_CODER_EXPONENT_START to ABC_MOST_SIGNIFICANT_BIT - 1 do
          Begin { For }
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := Decode_Mantissa_3;
                For Bit_Pos := ABC_RANGE_CODER_EXPONENT_START - 1 to Exponent - 1 do
                  Begin { For }
                    Mantissa := (Mantissa shl $01) or Decode_Mantissa;
                  End; { For }
                Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] := Mantissa;
                Inc (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
              End; { For }
          End; { For }

        { Save length }
        Compress_Information.P_Input_Record^.Size := M_Range_Coder.Done_Decoding;

        { Decode symbols }
        M_Symbol_Model := T_ABC_Symbol_Model.Create (ABC_BYTE_MAXIMUM_SYMBOL + 1 + ABC_MODEL_EOF_RANGE);
        M_Symbol_Model.Reset;
        M_EOF_Symbol := ABC_BYTE_MAXIMUM_SYMBOL + ABC_MODEL_EOF_RANGE;
        M_Range_Coder.Start_Decoding (Compress_Information.P_Input_Record^.P_Buffer, Compress_Information.P_Input_Record^.Size);
        Output_Index := 0;
        Symbol := Decode_Symbol;
        While Symbol <> M_EOF_Symbol do
          Begin { While }
            Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index] := Symbol;
            Inc (Output_Index);
            Symbol := Decode_Symbol;
          End; { While }
        Compress_Information.P_Input_Record^.Size   := M_Range_Coder.Done_Decoding;
        Compress_Information.P_Output_Record^.Size  := Output_Index;
        M_Symbol_Model.Free;

        { Set data for binary blocks }
        Compress_Information.P_Binary_Block^.A_Block_Length [Compress_Information.P_Binary_Block^.Block_Index]  := 0;
        { Reset A_Mantissa_Index }
        For I_0 := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
          Begin { For }
            Compress_Information.P_Binary_Block^.A_Mantissa_Index [I_0] := Compress_Information.P_Binary_Block^.A_Mantissa_Index [I_0] + Compress_Information.P_Binary_Block^.A_Frequency [I_0];
          End; { For }
      End { then }
        else
      Begin { else }
        If ABC_MCP.Compress_Information.SIF_F = TRUE then
          Begin { then }
            M_Byte_Model.Reset;
            M_Range_Coder.Start_Decoding (Compress_Information.P_Input_Record^.P_Buffer, Compress_Information.P_Input_Record^.Size);
            { Decode length of Exponent_Buffer }
            Block_Length := Decode_Byte;
            Block_Length := (Block_Length shl $08) or Decode_Byte;
            Block_Length := (Block_Length shl $08) or Decode_Byte;
            Compress_Information.P_Binary_Block^.Exponent_Index := Block_Length;
            { Decode length of Bit-Block }
            Block_Length := Decode_Byte;
            Block_Length := (Block_Length shl $08) or Decode_Byte;
            Block_Length := (Block_Length shl $08) or Decode_Byte;
            Compress_Information.P_Binary_Block^.Bit_Index := Block_Length;
            { Decode Block_Index }
            Block_Length := Decode_Byte;
            Block_Length := (Block_Length shl $08) or Decode_Byte;
            Block_Length := (Block_Length shl $08) or Decode_Byte;
            Compress_Information.P_Binary_Block^.Block_Index := Block_Length;
            { Decode length of binary blocks }
            For I := 0 to Compress_Information.P_Binary_Block^.Block_Index - 1 do
              Begin { For }
                Block_Length := Decode_Byte;
                Block_Length := (Block_Length shl $08) or Decode_Byte;
                Block_Length := (Block_Length shl $08) or Decode_Byte;
                Compress_Information.P_Binary_Block^.A_Block_Length [I] := Block_Length;
              End; { For }

            { Decode number of RLE-Blocks }
            FillChar (Compress_Information.P_Binary_Block^.A_RLE_Frequency [0], SizeOf (Compress_Information.P_Binary_Block^.A_RLE_Frequency), $00);
            Block_Length          := Decode_Byte;
            RLE_Frequency_Bit_Max := Block_Length;
            { Decode RLE-frequencies }
            For I := 0 to RLE_Frequency_Bit_Max do
              Begin { For }
                Block_Length := Decode_Byte;
                Block_Length := (Block_Length shl $08) or Decode_Byte;
                Block_Length := (Block_Length shl $08) or Decode_Byte;
                Compress_Information.P_Binary_Block^.A_RLE_Frequency [I] := Block_Length;
              End; { For }
            Compress_Information.P_Input_Record^.Size := M_Range_Coder.Done_Decoding;

            { Decode exponent }
            M_Exponent_Model.Reset;
            M_Range_Coder.Start_Decoding (Compress_Information.P_Input_Record^.P_Buffer, Compress_Information.P_Input_Record^.Size);
            For I := 0 to Compress_Information.P_Binary_Block^.Exponent_Index - 1 do
              Begin { For }
                Symbol := Decode_Exponent;
                Compress_Information.P_Binary_Block^.P_Exponent_Buffer^ [I] := Symbol;
              End; { For }
            Compress_Information.P_Input_Record^.Size := M_Range_Coder.Done_Decoding;

            { Calculate A_Frequency and A_Mantissa_Index }
            For I_0 := 0 to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                Compress_Information.P_Binary_Block^.A_Frequency [I_0]       := 0;
                Compress_Information.P_Binary_Block^.A_Mantissa_Index [I_0]  := 0;
              End; { For }
            For I_0 := 0 to Compress_Information.P_Binary_Block^.Exponent_Index - 1 do
              Begin { For }
                Exponent := Compress_Information.P_Binary_Block^.P_Exponent_Buffer^ [I_0];
                If Exponent > 0 then
                  Begin { then }
                    Inc (Compress_Information.P_Binary_Block^.A_Frequency [Exponent]);
                  End; { then }
              End; { For }
            For I_0 := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                Compress_Information.P_Binary_Block^.A_Mantissa_Index [I_0] := Compress_Information.P_Binary_Block^.A_Mantissa_Index [I_0 - 1] + Compress_Information.P_Binary_Block^.A_Frequency [I_0 - 1];
              End; { For }

            { Calculate A_RLE_Mantissa_Index }
            Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [0] := 0;
            For I := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I] := Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [I - 1] + Compress_Information.P_Binary_Block^.A_RLE_Frequency [I - 1];
              End; { For }

            { Create mantissa objects }
            M_Mantissa_Model.Reset;
            M_Mantissa_Model_1.Reset;
            M_Mantissa_Model_2.Reset;
            M_Mantissa_Model_3.Reset;
            M_Range_Coder.Start_Decoding (Compress_Information.P_Input_Record^.P_Buffer, Compress_Information.P_Input_Record^.Size);

            { Decode Mantisse_1 }
            Exponent := 1;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := Decode_Mantissa_1;
                Compress_Information.P_Binary_Block^.P_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent] + Mantissa_C] := Mantissa;
              End; { For }
            { Decode Mantisse_2 }
            Exponent := 2;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := Decode_Mantissa_2;
                Compress_Information.P_Binary_Block^.P_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent] + Mantissa_C] := Mantissa;
              End; { For }
            { Decode Mantisse_3 }
            Exponent := 3;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := Decode_Mantissa_3;
                Compress_Information.P_Binary_Block^.P_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent] + Mantissa_C] := Mantissa;
              End; { For }
            { Decode rest of mantissa bits }
            For Exponent := ABC_RANGE_CODER_EXPONENT_START to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_Frequency [Exponent] - 1 do
                  Begin { For }
                    Mantissa := Decode_Mantissa_3;
                    For Bit_Pos := ABC_RANGE_CODER_EXPONENT_START - 1 to Exponent - 1 do
                      Begin { For }
                        Mantissa := (Mantissa shl $01) or Decode_Mantissa;
                      End; { For }
                    Compress_Information.P_Binary_Block^.P_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent] + Mantissa_C] := Mantissa;
                  End; { For }
              End; { For }

            { Decode RLE-Mantisse_1 }
            Exponent := 1;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := Decode_Mantissa_1;
                Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] := Mantissa;
                Inc (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
              End; { For }
            { Decode RLE-Mantisse_2 }
            Exponent := 2;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := Decode_Mantissa_2;
                Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] := Mantissa;
                Inc (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
              End; { For }
            { Decode RLE-Mantisse_3 }
            Exponent := 3;
            For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
              Begin { For }
                Mantissa := Decode_Mantissa_3;
                Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] := Mantissa;
                Inc (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
              End; { For }
            { Decode rest of RLE mantissa bits }
            For Exponent := ABC_RANGE_CODER_EXPONENT_START to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                For Mantissa_C := 0 to Compress_Information.P_Binary_Block^.A_RLE_Frequency [Exponent] - 1 do
                  Begin { For }
                    Mantissa := Decode_Mantissa_3;
                    For Bit_Pos := ABC_RANGE_CODER_EXPONENT_START - 1 to Exponent - 1 do
                      Begin { For }
                        Mantissa := (Mantissa shl $01) or Decode_Mantissa;
                      End; { For }
                    Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] := Mantissa;
                    Inc (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
                  End; { For }
              End; { For }

            { Save length }
            Compress_Information.P_Input_Record^.Size := M_Range_Coder.Done_Decoding;

            { Initialize data for binary blocks }
            Compress_Information.P_Binary_Block^.A_Block_Length [Compress_Information.P_Binary_Block^.Block_Index]  := 0;
            { Reset A_Mantissa_Index }
            For I_0 := 1 to ABC_MOST_SIGNIFICANT_BIT - 1 do
              Begin { For }
                Compress_Information.P_Binary_Block^.A_Mantissa_Index [I_0] := Compress_Information.P_Binary_Block^.A_Mantissa_Index [I_0] + Compress_Information.P_Binary_Block^.A_Frequency [I_0];
              End; { For }
          End { then }
            else
          Begin { else }
            { Decode symbols }
            M_Symbol_Model := T_ABC_Symbol_Context_Model.Create (ABC_BYTE_MAXIMUM_SYMBOL + 1 + ABC_MODEL_EOF_RANGE);
            M_Symbol_Model.Reset;
            M_EOF_Symbol := ABC_BYTE_MAXIMUM_SYMBOL + ABC_MODEL_EOF_RANGE;
            M_Range_Coder.Start_Decoding (Compress_Information.P_Input_Record^.P_Buffer, Compress_Information.P_Input_Record^.Size);
            Output_Index := 0;
            Symbol := Decode_Symbol;
            While Symbol <> M_EOF_Symbol do
              Begin { While }
                Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index] := Symbol;
                Inc (Output_Index);
                Symbol := Decode_Symbol;
              End; { While }
            Compress_Information.P_Input_Record^.Size   := M_Range_Coder.Done_Decoding;
            Compress_Information.P_Output_Record^.Size  := Output_Index;
            M_Symbol_Model.Free;
          End; { else }
      End; { else }
  End; { T_ABC_Coder.Decode }


End.

