{#####################################################################################}
{##                                                                                 ##}
{## ABC - Advanced Blocksorting Compressor                                          ##}
{##                                                                                 ##}
{## ABC_MCP_Unit                                                                    ##}
{##                                                                                 ##}
{## Master Control Program : contains the main procedures of the program            ##}
{##                                                                                 ##}
{## Copyright (C) 2002-2003 Jrgen Abel                                             ##}
{##                                                                                 ##}
{#####################################################################################}


{$I ABC_Include.pas}


Unit ABC_MCP_Unit;


Interface


Uses
  Classes,
  SysUtils,
  ABC_Global_Unit,
  ABC_CRC_Unit,
  ABC_BWT_Unit,
  ABC_RLE_EXP_Unit,
  ABC_RLE_BIT_Unit,
  ABC_MTF_Unit,
  ABC_SIF_Unit,
  ABC_AWFC_Unit,
  ABC_Coder_Unit;


Const
  ABC_RLE_BIT_AWFC_ESCAPE_SYMBOL  = ABC_BYTE_MAXIMUM_SYMBOL;    { Escapesymbol for RLE_BIT }


Type
  T_ABC_MCP = Class
  Protected
    M_Mem_Block             : Pointer;
    M_Compress_Information  : T_R_ABC_Compress_Information;
    M_Input_File_Stream     : TFileStream;
    M_Output_File_Stream    : TFileStream;
    M_CRC32                 : T_ABC_CRC32;
    M_BWT                   : T_ABC_BWT;
    M_RLE_EXP               : T_ABC_RLE_EXP;
    M_RLE_BIT               : T_ABC_RLE_BIT;
    M_MTF                   : T_ABC_MTF;
    M_SIF                   : T_ABC_SIF;
    M_AWFC                  : T_ABC_AWFC;
    M_Coder                 : T_ABC_Coder;
    Function    Get_Compress_Information : T_R_ABC_Compress_Information;
  Public
    Constructor Create;
    Destructor  Destroy; Override;
    Procedure   Swap_Longword (Var L_0 : Longword; Var L_1 : Longword); Register;
    Procedure   Switch_Block (Var Compress_Information : T_R_ABC_Compress_Information);
    Function    Get_Bin_Length (Symbol : Integer) : Integer;
    Procedure   Write_Binary_Data (N : Integer);
    Function    Read_Binary_Data : Integer;
    Procedure   Push_Binary_Block;
    Procedure   Pop_Binary_Block;
    Procedure   Write_RLE_Data (N : Integer);
    Function    Read_RLE_Data (Exponent : Integer) : Integer;
    Procedure   Compress_Block;
    Procedure   Expand_Block;
    Procedure   Write_Compressed_Block;
    Procedure   Read_Compressed_Block;
    Procedure   Compress_File (Const Input_Filename : String; Const Output_Filename : String; Fast_F : Boolean);
    Procedure   Expand_File (Const Input_Filename : String; Const Output_Filename : String);
  Published
    Property    Compress_Information : T_R_ABC_Compress_Information Read Get_Compress_Information;
  End; { T_ABC_MCP }


Var
  ABC_MCP : T_ABC_MCP;


Implementation


Const
  ABC_ADMINISTRATION_LENGTH_SIZE  = 1;   { Bytesize of length of administration buffer }
  ABC_ADMINISTRATION_BITMASK_SIZE = 1;   { Bytesize of bitmask }
  ABC_ADMINISTRATION_FAST_F_BIT   = 0;   { Bitnumber for Fast_F in bitmask }
  ABC_ADMINISTRATION_SIF_F_BIT    = 1;   { Bitnumber for SIF_F in bitmask }
  ABC_RLE_BIT_AWFC_ROTATE_SYMBOL  = 0;   { Rotationsymbol fr WFC-RLE }


{-------------------------------------------------------------------------------------}
Constructor T_ABC_MCP.Create;
{ Allocates data                                                                      }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_MCP.Create }
    { Create records }
    M_Compress_Information.P_Input_Record         := New (T_P_R_ABC_IO_Block);
    M_Compress_Information.P_Output_Record        := New (T_P_R_ABC_IO_Block);
    M_Compress_Information.P_Administration_Block := New (T_P_R_ABC_Administration_Block);
    M_Compress_Information.P_Binary_Block         := New (T_P_R_ABC_Binary_Block);

    { Alloc buffers }
    GetMem (M_Mem_Block,
      ((ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE + (2 * ABC_BUFFER_FRONT_OFFSET)) * SizeOf (Byte)) +
      ((ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE + (2 * ABC_BUFFER_FRONT_OFFSET)) * SizeOf (Byte)) +
      (ABC_ADMINISTRATION_BUFFER_SIZE * SizeOf (Byte)) +
      ((ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE) * SizeOf (Byte)) +
      ((ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE) * SizeOf (Integer)) +
      ((ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE) * SizeOf (Integer)) +
      ((ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE) * SizeOf (Integer)) +
      ((ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE) * SizeOf (Integer))
    );

    { Set buffer pointers }
    M_Compress_Information.P_Input_Record^.P_Buffer                   := Pointer (Integer (M_Mem_Block));
    M_Compress_Information.P_Output_Record^.P_Buffer                  := Pointer (Integer (M_Compress_Information.P_Input_Record^.P_Buffer) + (ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE + (2 * ABC_BUFFER_FRONT_OFFSET)) * SizeOf (Byte));
    M_Compress_Information.P_Administration_Block^.P_Buffer           := Pointer (Integer (M_Compress_Information.P_Output_Record^.P_Buffer) + (ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE + (2 * ABC_BUFFER_FRONT_OFFSET)) * SizeOf (Byte));
    M_Compress_Information.P_Binary_Block^.P_Exponent_Buffer          := Pointer (Integer (M_Compress_Information.P_Administration_Block^.P_Buffer) + (ABC_ADMINISTRATION_BUFFER_SIZE * SizeOf (Byte)));
    M_Compress_Information.P_Binary_Block^.P_Mantissa_Buffer          := Pointer (Integer (M_Compress_Information.P_Binary_Block^.P_Exponent_Buffer) + ((ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE) * SizeOf (Byte)));
    M_Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer      := Pointer (Integer (M_Compress_Information.P_Binary_Block^.P_Mantissa_Buffer) + ((ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE) * SizeOf (Integer)));
    M_Compress_Information.P_Temp_A_0                                 := Pointer (Integer (M_Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer) + ((ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE) * SizeOf (Integer)));
    M_Compress_Information.P_Temp_A_1                                 := Pointer (Integer (M_Compress_Information.P_Temp_A_0) + ((ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE) * SizeOf (Integer)));
    Inc (Integer (Pointer (M_Compress_Information.P_Input_Record^.P_Buffer)), ABC_BUFFER_FRONT_OFFSET);
    Inc (Integer (Pointer (M_Compress_Information.P_Output_Record^.P_Buffer)), ABC_BUFFER_FRONT_OFFSET);


    { Create objects }
    M_CRC32 := T_ABC_CRC32.Create;
    M_BWT := T_ABC_BWT.Create;
    M_MTF := T_ABC_MTF.Create;
    M_AWFC := T_ABC_AWFC.Create;
    M_SIF := T_ABC_SIF.Create;
    M_RLE_EXP := T_ABC_RLE_EXP.Create;
    M_RLE_BIT := T_ABC_RLE_BIT.Create;
    M_Coder := T_ABC_Coder.Create;
  End; { T_ABC_MCP.Create }


{-------------------------------------------------------------------------------------}
Destructor T_ABC_MCP.Destroy;
{ Deallocates data                                                                    }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_MCP.Destroy }
    { Free objects }
    M_Coder.Free;
    M_RLE_BIT.Free;
    M_RLE_EXP.Free;
    M_SIF.Free;
    M_AWFC.Free;
    M_MTF.Free;
    M_BWT.Free;
    M_CRC32.Free;

    { Free buffers }
    FreeMem (M_Mem_Block);

    { Free records }
    Dispose (M_Compress_Information.P_Binary_Block);
    Dispose (M_Compress_Information.P_Administration_Block);
    Dispose (M_Compress_Information.P_Output_Record);
    Dispose (M_Compress_Information.P_Input_Record);
  End; { T_ABC_MCP.Destroy }


{-------------------------------------------------------------------------------------}
Function T_ABC_MCP.Get_Compress_Information : T_R_ABC_Compress_Information;
{ Returns the M_Compress_Information Record                                           }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_MCP.Get_Compress_Information }
    Result := M_Compress_Information;
  End; { T_ABC_MCP.Get_Compress_Information }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Swap_Longword (Var L_0 : Longword; Var L_1 : Longword); Register;
{ Switches 2 longwords                                                                }
{-------------------------------------------------------------------------------------}
  Var
    Temp : Longword;

  Begin { T_ABC_MCP.Swap_Longword }
    Temp  := L_0;
    L_0   := L_1;
    L_1   := Temp;
  End; { T_ABC_MCP.Swap_Longword }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Switch_Block (Var Compress_Information : T_R_ABC_Compress_Information);
{ Switches pointers for input and output block                                        }
{-------------------------------------------------------------------------------------}
  Var
    Temp : T_P_R_ABC_IO_Block;   { Zeiger auf Blockdatensatz }

  Begin { T_ABC_MCP.Switch_Block }
    Temp                                  := Compress_Information.P_Input_Record;
    Compress_Information.P_Input_Record   := Compress_Information.P_Output_Record;
    Compress_Information.P_Output_Record  := Temp;
  End; { T_ABC_MCP.Switch_Block }


{-------------------------------------------------------------------------------------}
Function T_ABC_MCP.Get_Bin_Length (Symbol : Integer) : Integer; Assembler;
{ Returns the binary length of the symbol                                             }
{-------------------------------------------------------------------------------------}
  {$IFDEF OPTIMIZATION_ON}
  Asm { T_ABC_MCP.Get_Bin_Length }
    INC     EDX
    BSR     EAX,EDX
  End; { T_ABC_MCP.Get_Bin_Length }
  {$ELSE}
  Var
    Bit_Index : Integer;
    Bit       : Integer;

  Begin { T_ABC_MCP.Get_Bin_Length }
    Inc (Symbol);
    Result := 0;
    For Bit_Index := ABC_MOST_SIGNIFICANT_BIT - 1 downto 0 do
      Begin { For }
        Bit := (Symbol shr Bit_Index) and $01;
        If Bit = 1 then
          Begin { then }
            Result := Bit_Index;
            Exit;
          End; { then }
      End; { For }
  End; { T_ABC_MCP.Get_Bin_Length }
  {$ENDIF}


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Write_Binary_Data (N : Integer);
{ Writes binary data                                                                  }
{-------------------------------------------------------------------------------------}
  Var
    Exponent : Byte;

  Begin { T_ABC_MCP.Write_Binary_Data }
    Exponent := Get_Bin_Length (N);

    { Save exponent }
    M_Compress_Information.P_Binary_Block^.P_Exponent_Buffer^ [M_Compress_Information.P_Binary_Block^.Exponent_Index] := Exponent;
    Inc (M_Compress_Information.P_Binary_Block^.Exponent_Index);

    If Exponent > 0 then
      Begin { then }
        { Increment exponent frequency }
        Inc (M_Compress_Information.P_Binary_Block^.A_Frequency [Exponent]);

        { Save mantissa }
        M_Compress_Information.P_Binary_Block^.P_Mantissa_Buffer^ [M_Compress_Information.P_Binary_Block^.Mantissa_Index] := (N + 1) - ($01 shl Exponent);
        Inc (M_Compress_Information.P_Binary_Block^.Mantissa_Index);
      End; { then }
  End; { T_ABC_MCP.Write_Binary_Data }


{-------------------------------------------------------------------------------------}
Function T_ABC_MCP.Read_Binary_Data : Integer;
{ Reads binary data                                                                   }
{-------------------------------------------------------------------------------------}
  Var
    Exponent : Byte;

  Begin { T_ABC_MCP.Read_Binary_Data }
    { Read exponent }
    Exponent := M_Compress_Information.P_Binary_Block^.P_Exponent_Buffer^ [M_Compress_Information.P_Binary_Block^.Exponent_Index];
    Inc (M_Compress_Information.P_Binary_Block^.Exponent_Index);

    { Read mantissa }
    If Exponent > 0 then
      Begin { then }
        Result := M_Compress_Information.P_Binary_Block^.P_Mantissa_Buffer^ [M_Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent]] + ($01 shl Exponent) - 1;
        Inc (M_Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent]);
      End { then }
        else
      Begin { else }
        Result := 0;
      End; { else }
  End; { T_ABC_MCP.Read_Binary_Data }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Push_Binary_Block;
{ Pushes indices for a binary block                                                   }
{-------------------------------------------------------------------------------------}
  Var
    N : Integer;

  Begin { T_ABC_MCP.Push_Binary_Block }
    { Get number of data }
    N                                                          := M_Compress_Information.P_Binary_Block^.Exponent_Index - M_Compress_Information.P_Binary_Block^.Last_Exponent_Index;
    M_Compress_Information.P_Binary_Block^.Last_Exponent_Index := M_Compress_Information.P_Binary_Block^.Exponent_Index;

    { Save number }
    M_Compress_Information.P_Binary_Block^.A_Block_Length [M_Compress_Information.P_Binary_Block^.Block_Index] := N;
    Inc (M_Compress_Information.P_Binary_Block^.Block_Index);
  End; { T_ABC_MCP.Push_Binary_Block }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Pop_Binary_Block;
{ Pops indices for a binary block                                                     }
{-------------------------------------------------------------------------------------}
  Procedure Reset_Binary_Block_Indices (N : Integer);
    Var
      I         : Integer;
      Exponent  : Integer;

    Begin { Reset_Binary_Block_Indices }
      For I := 0 to N - 1 do
        Begin { For }
          { Get exponent }
          Dec (M_Compress_Information.P_Binary_Block^.Exponent_Index);
          Exponent := M_Compress_Information.P_Binary_Block^.P_Exponent_Buffer^ [M_Compress_Information.P_Binary_Block^.Exponent_Index];

          If Exponent > 0 then
            Begin { then }
              { Reset index of mantissa }
              Dec (M_Compress_Information.P_Binary_Block^.A_Mantissa_Index [Exponent]);
            End; { then }
        End; { For }
    End; { Reset_Binary_Block_Indices }

  Var
    N : Integer;

  Begin { T_ABC_MCP.Pop_Binary_Block }
    { Reset data of current block }
    N := M_Compress_Information.P_Binary_Block^.A_Block_Length [M_Compress_Information.P_Binary_Block^.Block_Index];
    Reset_Binary_Block_Indices (N);

    { Reset data of former block }
    Dec (M_Compress_Information.P_Binary_Block^.Block_Index);
    N := M_Compress_Information.P_Binary_Block^.A_Block_Length [M_Compress_Information.P_Binary_Block^.Block_Index];
    Reset_Binary_Block_Indices (N);
  End; { T_ABC_MCP.Pop_Binary_Block }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Write_RLE_Data (N : Integer);
{ Writes RLE data                                                                     }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_MCP.Write_RLE_Data }
    M_Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [M_Compress_Information.P_Binary_Block^.RLE_Mantissa_Index] := N;
    Inc (M_Compress_Information.P_Binary_Block^.RLE_Mantissa_Index);
  End; { T_ABC_MCP.Write_RLE_Data }


{-------------------------------------------------------------------------------------}
Function T_ABC_MCP.Read_RLE_Data (Exponent : Integer) : Integer;
{ Reads RLE data                                                                      }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_MCP.Read_RLE_Data }
    { Get mantissa }
    If Exponent > 0 then
      Begin { then }
        Result := M_Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [M_Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] + ($01 shl Exponent) - 1;
        Inc (M_Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
      End { then }
        else
      Begin { else }
        Result := 0;
      End; { else }
  End; { T_ABC_MCP.Read_RLE_Data }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Compress_Block;
{ Compresses the input array                                                          }
{-------------------------------------------------------------------------------------}
  Var
    I                   : Integer;
    Symbol              : Integer;
    Run_Length          : Integer;
    Output_Index        : Integer;
    Bitnumber           : Integer;
    Mask                : Integer;
    Bit                 : Integer;

  Begin { T_ABC_MCP.Compress_Block }
    { Initialize data for binary blocks }
    M_Compress_Information.P_Binary_Block^.Exponent_Index       := 0;
    M_Compress_Information.P_Binary_Block^.Mantissa_Index       := 0;
    M_Compress_Information.P_Binary_Block^.RLE_Mantissa_Index   := 0;
    M_Compress_Information.P_Binary_Block^.Last_Exponent_Index  := 0;
    M_Compress_Information.P_Binary_Block^.Block_Index          := 0;
    For I := 0 to ABC_MOST_SIGNIFICANT_BIT - 1 do
      Begin { For }
        M_Compress_Information.P_Binary_Block^.A_Frequency [I]  := 0;
      End; { For }

    { Calculate and save CRC32 checksum }
    M_Compress_Information.CRC32_Compress := M_CRC32.Get_Buffer_CRC (M_Compress_Information.P_Input_Record^.P_Buffer, M_Compress_Information.P_Input_Record^.Size);
    M_Compress_Information.P_Administration_Block^.P_Buffer^ [M_Compress_Information.P_Administration_Block^.Cursor] := M_Compress_Information.CRC32_Compress and $FF;
    Inc (M_Compress_Information.P_Administration_Block^.Cursor);
    M_Compress_Information.P_Administration_Block^.P_Buffer^ [M_Compress_Information.P_Administration_Block^.Cursor] := (M_Compress_Information.CRC32_Compress shr 8) and $FF;
    Inc (M_Compress_Information.P_Administration_Block^.Cursor);
    M_Compress_Information.P_Administration_Block^.P_Buffer^ [M_Compress_Information.P_Administration_Block^.Cursor] := (M_Compress_Information.CRC32_Compress shr 16) and $FF;
    Inc (M_Compress_Information.P_Administration_Block^.Cursor);
    M_Compress_Information.P_Administration_Block^.P_Buffer^ [M_Compress_Information.P_Administration_Block^.Cursor] := (M_Compress_Information.CRC32_Compress shr 24) and $FF;
    Inc (M_Compress_Information.P_Administration_Block^.Cursor);

    { Check for fast operation }
    If M_Compress_Information.Fast_F = TRUE then
      Begin { then }
        { Fast operation, use MTF and no SIF or AWFC }
        M_Compress_Information.SIF_F := FALSE;

        M_BWT.Encode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);

        M_RLE_EXP.Threshold := ABC_RLE_EXP_IF_THRESHOLD;
        M_RLE_EXP.Encode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);

        M_MTF.Encode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);

        M_Coder.Encode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);
      End { then }
        else
      Begin { else }
        { Strong operation, use SIF or AWFC }
        M_BWT.Encode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);

        If M_Compress_Information.P_Input_Record.Size >= ABC_SIF_THRESHOLD then
          Begin { then }
            { Use SIF }
            M_Compress_Information.SIF_F := TRUE;

            M_RLE_EXP.Threshold := ABC_RLE_EXP_IF_THRESHOLD;
            M_RLE_EXP.Encode (M_Compress_Information);
            ABC_MCP.Switch_Block (M_Compress_Information);

            M_SIF.Encode (M_Compress_Information);
            ABC_MCP.Switch_Block (M_Compress_Information);
          End { then }
            else
          Begin { else }
            { Use AWFC }
            M_Compress_Information.SIF_F := FALSE;

            M_RLE_BIT.Threshold := ABC_RLE_BIT_AWFC_THRESHOLD;
            M_RLE_BIT.Encode (M_Compress_Information);
            ABC_MCP.Switch_Block (M_Compress_Information);

            M_AWFC.Encode (M_Compress_Information);
            ABC_MCP.Switch_Block (M_Compress_Information);

            Output_Index := 0;
            For I := 0 to M_Compress_Information.P_Input_Record^.Size - 1 do
              Begin { For }
                { Get symbol }
                Symbol := M_Compress_Information.P_Input_Record^.P_Buffer^ [I];

                { Rotate symbol and increment }
                If Symbol = ABC_RLE_BIT_AWFC_ROTATE_SYMBOL then
                  Begin { then }
                    Symbol := ABC_BYTE_MAXIMUM_SYMBOL + ABC_RLE_BIT_AWFC_SYMBOL_OFFSET;
                  End { then }
                    else
                  Begin { else }
                    Inc (Symbol, ABC_RLE_BIT_AWFC_SYMBOL_OFFSET - 1);
                  End; { else }

                { Encode symbol, if symbol >= ABC_BYTE_MAXIMUM_SYMBOL }
                If Symbol >= ABC_BYTE_MAXIMUM_SYMBOL then
                  Begin { then }
                    M_Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index] := ABC_RLE_BIT_AWFC_ESCAPE_SYMBOL;
                    Inc (Output_Index);
                    M_Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index] := Symbol - ABC_RLE_BIT_AWFC_ESCAPE_SYMBOL;
                    Inc (Output_Index);
                  End { then }
                    else
                  Begin { else }
                    M_Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index] := Symbol;
                    Inc (Output_Index);
                  End; { else }

                { Get run length }
                Run_Length := M_Compress_Information.P_Temp_A_0^ [I];

                { Encode run length }
                If Run_Length > 0 then
                  Begin { then }
                    Inc (Run_Length);

                    { Calcualte highest bit and decrement by 1 }
                    Bitnumber := ABC_MOST_SIGNIFICANT_BIT;
                    Mask      := $01 shl ABC_MOST_SIGNIFICANT_BIT;
                    While Run_Length and Mask = $00000000 do
                      Begin { While }
                        Mask := Mask shr $01;
                        Dec (Bitnumber);
                      End; { While }

                    { Output bits, LSB fist }
                    Mask := 1;
                    For Bit := 0 to Bitnumber - 1 do
                      Begin { For }
                        If Run_Length and Mask <> $00000000 then
                          Begin { then }
                            Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index] := ABC_RLE_BIT_AWFC_0B;
                            Inc (Output_Index);
                          End { then }
                            else
                          Begin { else }
                            Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index] := ABC_RLE_BIT_AWFC_0A;
                            Inc (Output_Index);
                          End; { else }

                        Mask := Mask shl 1;
                      End; { For }
                  End; { then }
              End; { For }
            M_Compress_Information.P_Output_Record^.Size := Output_Index;
            ABC_MCP.Switch_Block (M_Compress_Information);
          End; { else }

        M_Coder.Encode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);
      End; { else }

    ABC_MCP.Switch_Block (M_Compress_Information);
  End; { T_ABC_MCP.Compress_Block }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Expand_Block;
{ Expands the input array                                                             }
{-------------------------------------------------------------------------------------}
  Var
    Org_Input_Size      : Integer;
    Checksum            : Longword;
    Symbol              : Integer;
    Run_Length          : Integer;
    Input_Index         : Integer;
    Output_Index        : Integer;
    Mask                : Integer;

  Begin { T_ABC_MCP.Expand_Block }
    Org_Input_Size := 0;

    If M_Compress_Information.Fast_F = TRUE then
      Begin { then }
        M_Coder.Decode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);
        Inc (Org_Input_Size, M_Compress_Information.P_Output_Record^.Size);

        M_MTF.Decode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);

        M_RLE_EXP.Threshold := ABC_RLE_EXP_IF_THRESHOLD;
        M_RLE_EXP.Decode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);

        M_BWT.Decode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);
      End { then }
        else
      Begin { else }
        M_Coder.Decode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);
        Inc (Org_Input_Size, M_Compress_Information.P_Output_Record^.Size);

        If M_Compress_Information.SIF_F = TRUE then
          Begin { then }
            M_SIF.Decode (M_Compress_Information);
            ABC_MCP.Switch_Block (M_Compress_Information);

            M_RLE_EXP.Threshold := ABC_RLE_EXP_IF_THRESHOLD;
            M_RLE_EXP.Decode (M_Compress_Information);
            ABC_MCP.Switch_Block (M_Compress_Information);
          End { then }
            else
          Begin { else }
            Input_Index   := 0;
            Output_Index  := 0;
            While Input_Index <= M_Compress_Information.P_Input_Record^.Size - 1 do
              Begin { While }
                { Get symbol }
                Symbol := M_Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index];

                { Decode run length }
                If Symbol < ABC_RLE_BIT_AWFC_SYMBOL_OFFSET then
                  Begin { then }
                    { Calculate rest of run length }
                    Run_Length  := 0;
                    Mask        := $01;
                    Repeat
                      If Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index] = ABC_RLE_BIT_AWFC_0A then
                        Begin { then }
                          Inc (Run_Length, Mask);
                        End { then }
                          else
                        Begin { else }
                          Inc (Run_Length, (Mask shl 1));
                        End; { else }

                      Mask := Mask shl $01;

                      Inc (Input_Index);
                    until (Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index] > ABC_RLE_BIT_AWFC_0B) or (Input_Index >= Compress_Information.P_Input_Record^.Size);
                    { Output run length }
                    M_Compress_Information.P_Temp_A_0^ [Output_Index - 1] := Run_Length;
                  End { then }
                    else
                  Begin { else }
                    Inc (Input_Index);

                    { Encode Symbol, if symbol = ABC_BYTE_MAXIMUM_SYMBOL }
                    If Symbol = ABC_BYTE_MAXIMUM_SYMBOL then
                      Begin { then }
                        Symbol := Symbol + M_Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index];
                        Inc (Input_Index);
                      End; { then }

                    { Rotate symbol }
                    If Symbol = ABC_BYTE_MAXIMUM_SYMBOL + ABC_RLE_BIT_AWFC_SYMBOL_OFFSET then
                      Begin { then }
                        Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := ABC_RLE_BIT_AWFC_ROTATE_SYMBOL;
                        M_Compress_Information.P_Temp_A_0^ [Output_Index]      := 0;
                        Inc (Output_Index);
                      End { then }
                        else
                      Begin { else }
                        Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Symbol - (ABC_RLE_BIT_AWFC_SYMBOL_OFFSET - 1);
                        M_Compress_Information.P_Temp_A_0^ [Output_Index]      := 0;
                        Inc (Output_Index);
                      End; { else }
                  End; { else }
              End; { While }
            M_Compress_Information.P_Output_Record^.Size := Output_Index;
            ABC_MCP.Switch_Block (M_Compress_Information);

            M_AWFC.Decode (M_Compress_Information);
            ABC_MCP.Switch_Block (M_Compress_Information);

            M_RLE_BIT.Threshold := ABC_RLE_BIT_AWFC_THRESHOLD;
            M_RLE_BIT.Decode (M_Compress_Information);
            ABC_MCP.Switch_Block (M_Compress_Information);
          End; { else }

        M_BWT.Decode (M_Compress_Information);
        ABC_MCP.Switch_Block (M_Compress_Information);
      End; { else }
    { Calculate and restore CRC32 checksums }
    M_Compress_Information.CRC32_Expand := M_CRC32.Get_Buffer_CRC (M_Compress_Information.P_Input_Record^.P_Buffer, M_Compress_Information.P_Input_Record^.Size);
    Checksum := $00;
    Dec (M_Compress_Information.P_Administration_Block^.Cursor);
    Checksum := (Checksum shl 8) or M_Compress_Information.P_Administration_Block^.P_Buffer^ [M_Compress_Information.P_Administration_Block^.Cursor];
    Dec (M_Compress_Information.P_Administration_Block^.Cursor);
    Checksum := (Checksum shl 8) or M_Compress_Information.P_Administration_Block^.P_Buffer^ [M_Compress_Information.P_Administration_Block^.Cursor];
    Dec (M_Compress_Information.P_Administration_Block^.Cursor);
    Checksum := (Checksum shl 8) or M_Compress_Information.P_Administration_Block^.P_Buffer^ [M_Compress_Information.P_Administration_Block^.Cursor];
    Dec (M_Compress_Information.P_Administration_Block^.Cursor);
    Checksum := (Checksum shl 8) or M_Compress_Information.P_Administration_Block^.P_Buffer^ [M_Compress_Information.P_Administration_Block^.Cursor];
    M_Compress_Information.CRC32_Compress := Checksum;

    ABC_MCP.Switch_Block (M_Compress_Information);

    M_Compress_Information.P_Input_Record^.Size := Org_Input_Size;
  End; { T_ABC_MCP.Expand_Block }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Write_Compressed_Block;
{ Writes a compressed block to the filestream                                         }
{-------------------------------------------------------------------------------------}
  Var
    Bitmask : Byte;
    Cursor  : Byte;

  Begin { T_ABC_MCP.Write_Compressed_Block }
    { Set length of administration buffer }
    Cursor                                                            := 0;
    M_Compress_Information.P_Administration_Block^.P_Buffer^ [Cursor] := M_Compress_Information.P_Administration_Block^.Cursor and $FF;
    Inc (Cursor);

    { Set bitmask }
    Bitmask := $00;
    If M_Compress_Information.Fast_F = TRUE then
      Begin { then }
        Bitmask := Bitmask or (1 shl ABC_ADMINISTRATION_FAST_F_BIT);
      End; { then }
    If M_Compress_Information.SIF_F = TRUE then
      Begin { then }
        Bitmask := Bitmask or (1 shl ABC_ADMINISTRATION_SIF_F_BIT);
      End; { then }
    M_Compress_Information.P_Administration_Block^.P_Buffer^ [Cursor] := Bitmask;

    { Write administration buffer }
    M_Output_File_Stream.Write (M_Compress_Information.P_Administration_Block^.P_Buffer^, M_Compress_Information.P_Administration_Block^.Cursor);

    { Write output buffer }
    M_Output_File_Stream.Write (M_Compress_Information.P_Output_Record^.P_Buffer^, M_Compress_Information.P_Output_Record^.Size);
  End; { T_ABC_MCP.Write_Compressed_Block }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Read_Compressed_Block;
{ Reads a compressed block from the filestream                                        }
{-------------------------------------------------------------------------------------}
  Var
    Bitmask : Byte;

  Begin { T_ABC_MCP.Read_Compressed_Block }
    { Read administration buffer }
    M_Input_File_Stream.Read (M_Compress_Information.P_Administration_Block^.P_Buffer^, ABC_ADMINISTRATION_LENGTH_SIZE);
    M_Compress_Information.P_Administration_Block^.Cursor := M_Compress_Information.P_Administration_Block^.P_Buffer^ [0];
    M_Input_File_Stream.Read (T_P_ABC_Byte_Array (Longword (M_Compress_Information.P_Administration_Block^.P_Buffer) + ABC_ADMINISTRATION_LENGTH_SIZE)^, M_Compress_Information.P_Administration_Block^.Cursor - ABC_ADMINISTRATION_LENGTH_SIZE);

    { Set flags }
    Bitmask := M_Compress_Information.P_Administration_Block^.P_Buffer^ [ABC_ADMINISTRATION_LENGTH_SIZE];
    If Bitmask and (1 shl ABC_ADMINISTRATION_FAST_F_BIT) <> $00 then
      Begin { then }
        M_Compress_Information.Fast_F := TRUE;
      End { then }
        else
      Begin { else }
        M_Compress_Information.Fast_F := FALSE;
      End; { else }
    If Bitmask and (1 shl ABC_ADMINISTRATION_SIF_F_BIT) <> $00 then
      Begin { then }
        M_Compress_Information.SIF_F := TRUE;
      End { then }
        else
      Begin { else }
        M_Compress_Information.SIF_F := FALSE;
      End; { else }

    { Read input buffer }
    M_Compress_Information.P_Input_Record^.Size := M_Input_File_Stream.Read (M_Compress_Information.P_Input_Record^.P_Buffer^, ABC_BUFFER_SIZE + ABC_BUFFER_SAVETY_SIZE);
  End; { T_ABC_MCP.Read_Compressed_Block }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Compress_File (Const Input_Filename : String; Const Output_Filename : String; Fast_F : Boolean);
{ Compresses a file                                                                   }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_MCP.Compress_File }
    M_Compress_Information.Fast_F := Fast_F;

    M_Input_File_Stream := TFileStream.Create (Input_Filename, fmOpenRead or fmShareDenyWrite);
    Try
      M_Output_File_Stream := TFileStream.Create (Output_Filename, fmCreate or fmShareExclusive);
      Try
        { Read first block }
        M_Compress_Information.P_Input_Record^.Size := M_Input_File_Stream.Read (M_Compress_Information.P_Input_Record^.P_Buffer^, ABC_BUFFER_SIZE);

        { Compress file in blocks }
        While (M_Compress_Information.P_Input_Record^.Size > 0) do
          Begin { While }
            { Compress a block }
            M_Compress_Information.P_Administration_Block^.Cursor := ABC_ADMINISTRATION_LENGTH_SIZE + ABC_ADMINISTRATION_BITMASK_SIZE;
            Compress_Block;

            { Write block }
            Write_Compressed_Block;

            { Read next block }
            M_Compress_Information.P_Input_Record^.Size := M_Input_File_Stream.Read (M_Compress_Information.P_Input_Record^.P_Buffer^, ABC_BUFFER_SIZE);
          End; { While }
      Finally
        M_Output_File_Stream.Free;
      End; { Try }
    Finally
      M_Input_File_Stream.Free;
    End; { Try }
  End; { T_ABC_MCP.Compress_File }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MCP.Expand_File (Const Input_Filename : String; Const Output_Filename : String);
{ Expands a file                                                                      }
{-------------------------------------------------------------------------------------}
  Var
    Offset                    : Integer;
    Administration_Block_Size : Integer;

  Begin { T_ABC_MCP.Expand_File }
    M_Input_File_Stream := TFileStream.Create (Input_Filename, fmOpenRead or fmShareDenyWrite);
    Try
      M_Output_File_Stream := TFileStream.Create (Output_Filename, fmCreate or fmShareExclusive);
      Try
        { Read first block }
        Read_Compressed_Block;
        Administration_Block_Size := M_Compress_Information.P_Administration_Block^.Cursor;
        Offset                    := 0;

        { Process file in blocks }
        While (M_Compress_Information.P_Input_Record^.Size > 0) do
          Begin { While }
            { Expand block }
            Expand_Block;

            { Check CRC }
            If M_Compress_Information.CRC32_Compress <> M_Compress_Information.CRC32_Expand then
              Begin { then }
                WriteLn (ABC_IO_EMPTY_TEXT);
                WriteLn (Format (ABC_IO_CRC_NOT_VALID, []));
                {$IFNDEF NO_WAIT}
                { Wait for RETURN }
                ReadLn;
                {$ENDIF}
                Halt;
              End; { then }

            { Write block }
            M_Output_File_Stream.Write (M_Compress_Information.P_Output_Record^.P_Buffer^, M_Compress_Information.P_Output_Record^.Size);

            { Read next block }
            Offset := Offset + Administration_Block_Size + M_Compress_Information.P_Input_Record^.Size;
            M_Input_File_Stream.Seek (Offset, soFromBeginning);
            Read_Compressed_Block;
            Administration_Block_Size := M_Compress_Information.P_Administration_Block^.Cursor;
          End; { While }
      Finally
        M_Output_File_Stream.Free;
      End; { Try }
    Finally
      M_Input_File_Stream.Free;
    End; { Try }
  End; { T_ABC_MCP.Expand_File }


End.

