{#####################################################################################}
{##                                                                                 ##}
{## ABC - Advanced Blocksorting Compressor                                          ##}
{##                                                                                 ##}
{## ABC_Model_Unit                                                                  ##}
{##                                                                                 ##}
{## Adaptive models for the arithmetic coder                                        ##}
{##                                                                                 ##}
{## Copyright (C) 2002-2003 Jrgen Abel                                             ##}
{##                                                                                 ##}
{#####################################################################################}


{$I ABC_Include.pas}


Unit ABC_Model_Unit;


Interface


Uses
  ABC_Global_Unit;


Const
  ABC_MODEL_FIRST_LEVEL_SYMBOL_N                = 4;                                       { Maximum symbol for 1. level model }
  ABC_MODEL_SECOND_LEVEL_SYMBOL_N               = 8;                                       { Maximum symbol for 2. level model }
  ABC_MODEL_FIRST_LEVEL_ESCAPE_SYMBOL           = ABC_MODEL_FIRST_LEVEL_SYMBOL_N - 1;      { Escape symbol for 1. level model }
  ABC_MODEL_RESCALE_COUNT                       = 1;
  ABC_MODEL_EXPONENT_FIRST_LEVEL_SYMBOL_N       = 5;                                       { Maximum symbol for 2. level model }
  ABC_MODEL_EOF_RANGE                           = 1;                                       { Range for eof symbol }
  ABC_MODEL_SYMBOL_MAX                          = ABC_BYTE_MAXIMUM_SYMBOL + ABC_RLE_BIT_AWFC_SYMBOL_OFFSET + ABC_MODEL_EOF_RANGE;  { Maximum symbol }
  ABC_MODEL_FREQUENCY_MAX                       = 32*ABC_KB;                               { maximum frequency of the models }
  ABC_MODEL_INIT_FACTOR                         = 1;                                       { Factor for initialising the frequencies of the model }
  ABC_MODEL_CONTEXT_MAXIMUM_ADRESS_MASK         = $03FF;                                   { Adressmask for contexts }

Type
  T_ABC_Model_A_Third_Level_Integer             = Array [0 .. ABC_MODEL_SECOND_LEVEL_SYMBOL_N - 1] of Integer;                               { Array of integer for increment }
  T_ABC_Model_A_Third_Level_Start_Integer       = Array [0 .. ABC_MODEL_SECOND_LEVEL_SYMBOL_N] of Integer;                                   { Array of integer for group start }
  T_ABC_Model_A_Second_Level_Integer            = Array [0 .. ABC_MODEL_EXPONENT_FIRST_LEVEL_SYMBOL_N - 1] of Integer;                       { Array of integer for increment }
  T_ABC_Model_A_Second_Level_Start_Integer      = Array [0 .. ABC_MODEL_EXPONENT_FIRST_LEVEL_SYMBOL_N] of Integer;                           { Array of integer for group start }
  T_ABC_Model_A_Context_Frequency_0             = Array [0 .. ABC_MODEL_SYMBOL_MAX + 1] of Integer;                                          { Array for context-0 frequencies }
  T_ABC_Model_A_Context_Frequency_n             = Array [0 .. ABC_MODEL_CONTEXT_MAXIMUM_ADRESS_MASK, 0 .. ABC_MODEL_SYMBOL_MAX + 1] of Word; { Array for context-n frequencies }
  T_P_ABC_Model_A_Context_Frequency_0           = ^T_ABC_Model_A_Context_Frequency_0;
  T_P_ABC_Model_A_Context_Frequency_n           = ^T_ABC_Model_A_Context_Frequency_n;

Const
  ABC_MODEL_FIRST_LEVEL_INCREMENT_0             = 900;    { Increment for 1. level order-0 }
  ABC_MODEL_FIRST_LEVEL_INCREMENT_1             = 400;    { Increment for 1. level order-1 }
  ABC_MODEL_FIRST_LEVEL_INCREMENT_n             = 500;    { Increment for 1. level order-n }

  ABC_MODEL_FIRST_LEVEL_CONTEXT_BIT_SHIFT       = 2;      { Bitshift for 1. level order-n context }
  ABC_MODEL_FIRST_LEVEL_n                       = 5;      { Value of n for 1. level }


  ABC_MODEL_SECOND_LEVEL_INCREMENT_0            = 1000;   { Increment for 2. level order-0 }
  ABC_MODEL_SECOND_LEVEL_INCREMENT_1            = 500;    { Increment for 2. level order-1 }
  ABC_MODEL_SECOND_LEVEL_INCREMENT_n            = 150;    { Increment for 2. level order-n }

  ABC_MODEL_SECOND_LEVEL_CONTEXT_BIT_SHIFT      = 3;      { Bitshift for 2. level order-n context }
  ABC_MODEL_SECOND_LEVEL_n                      = 2;      { Value of n for 2. level }


  ABC_MODEL_A_THIRD_LEVEL_INCREMENT_0           : T_ABC_Model_A_Third_Level_Integer =
    (0, 128, 16, 16, 16, 8, 8, 8);                        { Increment for 3. level order-0 }

  ABC_MODEL_A_THIRD_LEVEL_GROUP_START           : T_ABC_Model_A_Third_Level_Start_Integer =
    ( 3, 4, 6, 10, 16, 32, 70, 150, ABC_MODEL_SYMBOL_MAX + 1);  { Group start for 3. level }


  ABC_MODEL_EXPONENT_INCREMENT_0                = 1400;   { Increment for exponent order-0 }
  ABC_MODEL_EXPONENT_INCREMENT_1                = 400;    { Increment for exponent order-1 }
  ABC_MODEL_EXPONENT_INCREMENT_n                = 200;    { Increment for exponent order-n }

  ABC_MODEL_EXPONENT_CONTEXT_BIT_SHIFT          = 3;      { Bitshift for exponent order-n context }
  ABC_MODEL_EXPONENT_n                          = 3;      { Value of n for exponent }

  ABC_MODEL_A_SECOND_LEVEL_INCREMENT_0          : T_ABC_Model_A_Second_Level_Integer =
    (0, 0, 0, 0, 1000);                                   { Increment for 2. level order-0 }

  ABC_MODEL_A_SECOND_LEVEL_GROUP_START          : T_ABC_Model_A_Second_Level_Start_Integer =
    (0, 1, 2, 3, 4, ABC_MOST_SIGNIFICANT_BIT);            { Group start for 2. level }


  ABC_MODEL_MANTISSA_INCREMENT_0                = 1;      { Increment for mantissa order-0 }
  ABC_MODEL_MANTISSA_INCREMENT_1                = 1;      { Increment for mantissa order-1 }
  ABC_MODEL_MANTISSA_INCREMENT_n                = 1;      { Increment for mantissa order-n }

  ABC_MODEL_MANTISSA_CONTEXT_BIT_SHIFT          = 1;      { Bitshift for mantissa order-n context }
  ABC_MODEL_MANTISSA_n                          = 5;      { Value of n for mantissa }


Type
  T_ABC_Model = Class
  Protected
    M_Increment             : Integer;                                          { Increment }
    M_A_Frequency           : Array [0 .. ABC_MODEL_SYMBOL_MAX + 1] of Integer; { Descending sorted array of frequencies }
    M_Number_Of_Symbols     : Integer;                                          { Symbol number }
  Public
    M_A_Symbol_To_Index     : Array [0 .. ABC_MODEL_SYMBOL_MAX + 1] of Integer; { Lookup table }
    M_A_Index_To_Symbol     : Array [0 .. ABC_MODEL_SYMBOL_MAX + 1] of Integer; { Lookup table }
    M_A_Cumulated_Frequency : Array [0 .. ABC_MODEL_SYMBOL_MAX + 1] of Integer; { Descending sorted array of cumulated frequencies }
    Constructor Create (Number_Of_Symbols : Integer; Increment : Integer);
    Destructor  Destroy; Override;
    Function    Index_To_Symbol (Index : Integer) : Integer;
    Procedure   Reset; Virtual;
    Procedure   Update (Index : Integer); Virtual;
    Property    Number_Of_Symbols : Integer Read M_Number_Of_Symbols;
  End; { T_ABC_Model }

  T_ABC_Context_Model = Class (T_ABC_Model)
  Protected
    M_Cache                               : Longword; { Last data }
    M_Context_Bit_Shift                   : Integer;  { Number of bits to shift }
    M_n                                   : Integer;  { Order n }
    M_Context_1                           : Longword; { Current context 1 }
    M_Context_1_Mask                      : Longword; { Context 1 mask }
    M_Context_n                           : Longword; { Current context n }
    M_Context_n_Mask                      : Longword; { Context n mask }
    M_Increment_0                         : Integer;  { Increment order 0 }
    M_Increment_1                         : Integer;  { Increment order 1 }
    M_Increment_n                         : Integer;  { Increment order n }
    M_Increment_C                         : Integer;  { Increment counter }
    M_Old_Index                           : Integer;  { Old index }
    M_Mem_Block                           : Pointer;  { Main memory block }
    M_P_A_Context_Frequency_0             : T_P_ABC_Model_A_Context_Frequency_0; { Frequency array order 0 }
    M_P_A_Context_Cumulated_Frequency_0   : T_P_ABC_Model_A_Context_Frequency_0; { Cumulated frequency array order 0 }
    M_P_A_Context_Frequency_1             : T_P_ABC_Model_A_Context_Frequency_n; { Frequency array order 1 }
    M_P_A_Context_Cumulated_Frequency_1   : T_P_ABC_Model_A_Context_Frequency_n; { Cumulated frequency array order 1 }
    M_P_A_Context_Frequency_n             : T_P_ABC_Model_A_Context_Frequency_n; { Frequency array order n }
    M_P_A_Context_Cumulated_Frequency_n   : T_P_ABC_Model_A_Context_Frequency_n; { Cumulated frequency array order n }
    Procedure   Set_Current_Frequencies (Index : Integer);
    Procedure   Update_Context (Index : Integer); Virtual;
  Public
    Constructor Create (Number_Of_Symbols : Integer; Increment_0 : Integer; Increment_1 : Integer; Increment_n : Integer; Context_Bit_Shift : Integer; n : Integer);
    Destructor  Destroy; Override;
    Procedure   Reset; Override;
    Procedure   Update (Index : Integer); Override;
  End; { T_ABC_Model }

  T_ABC_Context_Level_0_Model = Class (T_ABC_Context_Model)
  Public
    Procedure   Update_Context (Index : Integer); Override;
  End; { T_ABC_Model }

  T_A_Third_Level_Model   = Array [0 .. ABC_MODEL_SECOND_LEVEL_SYMBOL_N - 1] of T_ABC_Model; { Array of 3. level models }

  T_ABC_Symbol_Model = Class
  Protected
    M_Alphabet_Size                         : Integer;                                      { Alphabet size }
    Function    Get_Third_Level_Model (I : Integer) : T_ABC_Model;                          { Returns 3. level model }
  Public
    M_First_Level_Model                     : T_ABC_Model;                                  { 1. level model }
    M_Second_Level_Model                    : T_ABC_Model;                                  { 2. level model }
    M_A_Third_Level_Model                   : T_A_Third_Level_Model;                        { Array of 3. level models }
    M_A_Symbol_To_Second_Level_Group_Number : Array [0 .. ABC_MODEL_SYMBOL_MAX] of Integer; { Lookup table }
    M_Third_Level_Model_N                   : Integer;                                      { Number of 3. level models }
    Constructor Create (Alphabet_Size : Integer);
    Destructor  Destroy; Override;
    Function    Third_Level_Group_Offset_To_Symbol (Group_Number : Integer; Group_Index : Integer) : Integer;
    Procedure   Reset;
    Procedure   Get_Hierarchical_Information (Symbol : Integer; Var Model_0 : T_ABC_Model; Var Symbol_0 : Integer; Var Model_1 : T_ABC_Model; Var Symbol_1 : Integer; Var Model_2 : T_ABC_Model; Var Symbol_2 : Integer);
    Property    First_Level_Model : T_ABC_Model                  Read M_First_Level_Model;
    Property    Second_Level_Model : T_ABC_Model                 Read M_Second_Level_Model;
    Property    A_Third_Level_Model [I : Integer] : T_ABC_Model  Read Get_Third_Level_Model;
  End; { T_ABC_Symbol_Model }

  T_ABC_Symbol_Context_Model = Class (T_ABC_Symbol_Model)
  Protected
  Public
    Constructor Create (Alphabet_Size : Integer);
  End; { T_ABC_Symbol_Context_Model }

  T_A_Second_Level_Model   = Array [0 .. ABC_MOST_SIGNIFICANT_BIT - 1] of T_ABC_Model; { Array of 2. level models }

  T_ABC_Exponent_Model = Class
  Protected
    Function    Get_Second_Level_Model (I : Integer) : T_ABC_Model;
  Public
    M_First_Level_Model                     : T_ABC_Model;                                          { 1. level model }
    M_A_Second_Level_Model                  : T_A_Third_Level_Model;                                { Array of 3. level models }
    M_A_Symbol_To_Second_Level_Group_Number : Array [0 .. ABC_MOST_SIGNIFICANT_BIT - 1] of Integer; { Transformation array }
    Constructor Create;
    Destructor  Destroy; Override;
    Function    Second_Level_Group_Offset_To_Symbol (Group_Number : Integer; Group_Index : Integer) : Integer;
    Procedure   Reset;
    Procedure   Get_Hierarchical_Information (Symbol : Integer; Var Model_0 : T_ABC_Model; Var Symbol_0 : Integer; Var Model_1 : T_ABC_Model; Var Symbol_1 : Integer);
    Property    First_Level_Model : T_ABC_Model                  Read M_First_Level_Model;
    Property    A_Second_Level_Model [I : Integer] : T_ABC_Model Read Get_Second_Level_Model;
  End; { T_ABC_Symbol_Model }

  T_ABC_Mantissa_Model = Class
  Protected
    M_Model : T_ABC_Model;
  Public
    Constructor Create;
    Destructor  Destroy; Override;
    Procedure   Reset;
    Property    Model : T_ABC_Model Read M_Model;
  End; { T_ABC_Mantissa_Model }

  T_ABC_Mantissa_Model_1 = Class (T_ABC_Mantissa_Model)
  Public
    Constructor Create;
  End; { T_ABC_Mantissa_Model_1 }

  T_ABC_Mantissa_Model_2 = Class (T_ABC_Mantissa_Model)
  Public
    Constructor Create;
  End; { T_ABC_Mantissa_Model_2 }

  T_ABC_Mantissa_Model_3 = Class (T_ABC_Mantissa_Model)
  Public
    Constructor Create;
  End; { T_ABC_Mantissa_Model_3 }


Implementation


Uses
  ABC_MCP_Unit;


{#####################################################################################}
{ T_ABC_Model                                                                      }
{#####################################################################################}


{-------------------------------------------------------------------------------------}
Constructor T_ABC_Model.Create (Number_Of_Symbols : Integer; Increment : Integer);
{ Initializes members                                                                 }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Model.Create }
    { Save variables }
    M_Number_Of_Symbols := Number_Of_Symbols;
    M_Increment         := Increment;
  End; { T_ABC_Model.Create }


{-------------------------------------------------------------------------------------}
Destructor T_ABC_Model.Destroy;
{ Frees data                                                                          }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Model.Destroy }
    Inherited Destroy;
  End; { T_ABC_Model.Destroy }


{-------------------------------------------------------------------------------------}
Function T_ABC_Model.Index_To_Symbol (Index : Integer) : Integer;
{ Lookup from index to symbol                                                         }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Model.Index_To_Symbol }
    Result := M_A_Index_To_Symbol [Index];
  End; { T_ABC_Model.Index_To_Symbol }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Model.Reset;
{ Resets frequency table                                                              }
{-------------------------------------------------------------------------------------}
  Var
    I                   : Integer;
    Cumulated_Frequency : Integer;

  Begin { T_ABC_Model.Reset }
    { Initialize lookup tables }
    For I := 0 To M_Number_Of_Symbols - 1 do
      Begin { For }
        M_A_Symbol_To_Index [I]     := I + 1;
        M_A_Index_To_Symbol [I + 1] := I;
      End; { For }

    { Frequency of 0 must be 0 }
    M_A_Frequency [0] := 0;

    { Initialize frequency array }
    For I := 1 to M_Number_Of_Symbols do
      Begin { For }
        M_A_Frequency [I] := (M_Number_Of_Symbols + 1 - I) * M_Increment * ABC_MODEL_INIT_FACTOR;
      End; { For }

    { Initialize cumulated frequency array }
    Cumulated_Frequency := 0;
    For I := M_Number_Of_Symbols downto 0 do
      Begin { For }
        M_A_Cumulated_Frequency [I] := Cumulated_Frequency;
        Inc (Cumulated_Frequency, M_A_Frequency [I]);
      End; { For }
  End; { T_ABC_Model.Reset }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Model.Update (Index : Integer);
{ Updates the model                                                                   }
{ Frequency array and cumulated frequency array are decending sorted                  }
{-------------------------------------------------------------------------------------}
  Var
    I                   : Integer;
    Old_Symbol          : Integer;
    New_Symbol          : Integer;
    Cumulated_Frequency : Integer;

  Begin { T_ABC_Model.Update }
    { Calculate new position for index }
    I := Index;
    While M_A_Frequency [I - 1] = M_A_Frequency [I] do
      Begin { While }
        Dec (I);
      End; { While }
    { Update data, if position changed }
    If I < Index then
      Begin { then }
        Old_Symbol                        := M_A_Index_To_Symbol [I];
        New_Symbol                        := M_A_Index_To_Symbol [Index];

        M_A_Index_To_Symbol [I]           := New_Symbol;
        M_A_Index_To_Symbol [Index]       := Old_Symbol;

        M_A_Symbol_To_Index [Old_Symbol]    := Index;
        M_A_Symbol_To_Index [New_Symbol]    := I;
      End; { then }

    { Increment frequency }
    Inc (M_A_Frequency [I], M_Increment);

    { Update frequencies on top }
    While I > 0 do
      Begin { While }
        Dec (I);
        Inc (M_A_Cumulated_Frequency [I], M_Increment);
      End; { While }

    { Rescale if cumulated frequency passes ABC_MODEL_FREQUENCY_MAX }
    If M_A_Cumulated_Frequency [0] >= ABC_MODEL_FREQUENCY_MAX then
      Begin { then }
        Cumulated_Frequency := 0;
        For I := M_Number_Of_Symbols downto 0 do
          Begin { For }
            M_A_Frequency [I]         := (M_A_Frequency [I] + 1) shr 1;
            M_A_Cumulated_Frequency [I] := Cumulated_Frequency;
            Inc (Cumulated_Frequency, M_A_Frequency [I]);
          End; { For }
      End; { then }
  End; { T_ABC_Model.Update }

 
{#####################################################################################}
{ T_ABC_Context_Model                                                              }
{#####################################################################################}


{-------------------------------------------------------------------------------------}
Constructor T_ABC_Context_Model.Create (Number_Of_Symbols : Integer; Increment_0 : Integer; Increment_1 : Integer; Increment_n : Integer; Context_Bit_Shift : Integer; n : Integer);
{ Initializes members and allocates memory                                            }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;

  Begin { T_ABC_Context_Model.Create }
    { Allocate main memory block }
    GetMem (M_Mem_Block,
      SizeOf (T_ABC_Model_A_Context_Frequency_0) +
      SizeOf (T_ABC_Model_A_Context_Frequency_0) +
      SizeOf (T_ABC_Model_A_Context_Frequency_n) +
      SizeOf (T_ABC_Model_A_Context_Frequency_n) +
      SizeOf (T_ABC_Model_A_Context_Frequency_n) +
      SizeOf (T_ABC_Model_A_Context_Frequency_n)
    );

    M_P_A_Context_Frequency_0           := M_Mem_BLock;
    M_P_A_Context_Cumulated_Frequency_0 := Pointer (Integer (M_P_A_Context_Frequency_0) + SizeOf (T_ABC_Model_A_Context_Frequency_0));
    M_P_A_Context_Frequency_1           := Pointer (Integer (M_P_A_Context_Cumulated_Frequency_0) + SizeOf (T_ABC_Model_A_Context_Frequency_0));
    M_P_A_Context_Cumulated_Frequency_1 := Pointer (Integer (M_P_A_Context_Frequency_1) + SizeOf (T_ABC_Model_A_Context_Frequency_n));
    M_P_A_Context_Frequency_n           := Pointer (Integer (M_P_A_Context_Cumulated_Frequency_1) + SizeOf (T_ABC_Model_A_Context_Frequency_n));
    M_P_A_Context_Cumulated_Frequency_n := Pointer (Integer (M_P_A_Context_Frequency_n) + SizeOf (T_ABC_Model_A_Context_Frequency_n));


    { Save variables }
    M_Context_Bit_Shift   := Context_Bit_Shift;
    M_n                   := n;
    M_Increment_0         := Increment_0;
    M_Increment_1         := Increment_1;
    M_Increment_n         := Increment_n;

    M_Context_1_Mask := $01;
    For I := 1 to M_Context_Bit_Shift - 1 do
      Begin { For }
        M_Context_1_Mask := (M_Context_1_Mask shl $01) or $01;
      End; { For }

    M_Context_n_Mask := M_Context_1_Mask;
    For I := 1 to n - 1 do
      Begin { For }
        M_Context_n_Mask := (M_Context_n_Mask shl M_Context_Bit_Shift) or M_Context_1_Mask;
      End; { For }

    Inherited Create (Number_Of_Symbols, Increment_0);
  End; { T_ABC_Context_Model.Create }
 

{-------------------------------------------------------------------------------------}
Destructor T_ABC_Context_Model.Destroy;
{ Frees memory                                                                        }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Context_Model.Destroy }
    { Free memory }
    FreeMem (M_Mem_Block);

    Inherited Destroy;
  End; { T_ABC_Context_Model.Destroy }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Context_Model.Set_Current_Frequencies (Index : Integer);
{ Calculates current frequency array                                                  }
{-------------------------------------------------------------------------------------}
  Var
    I                   : Integer;
    Cumulated_Frequency : Integer;

  Begin {T_ABC_Context_Model.Set_Current_Frequencies }
    { Calculate current frequency array }
    If ABC_MCP.Compress_Information.SIF_F = TRUE then
      Begin { then }
        Cumulated_Frequency := 0;
        For I := M_Number_Of_Symbols downto 0 do
          Begin { For }
            M_A_Cumulated_Frequency [I] := Cumulated_Frequency;
            Inc (Cumulated_Frequency, (((M_P_A_Context_Frequency_0^ [I]) + (M_P_A_Context_Frequency_1^ [M_Context_1, I] shr 1) + (M_P_A_Context_Frequency_n^ [M_Context_n, I])) shr 2) + 1);
          End; { For }
      End { then }
        else
      Begin { else }
        Cumulated_Frequency := 0;
        For I := M_Number_Of_Symbols downto 0 do
          Begin { For }
            M_A_Cumulated_Frequency [I] := Cumulated_Frequency;
            Inc (Cumulated_Frequency, ((M_P_A_Context_Frequency_0^ [I] + M_P_A_Context_Frequency_1^ [M_Context_1, I] + (M_P_A_Context_Frequency_n^ [M_Context_n, I] shl 1)) shr 2) + 1);
          End; { For }
      End; { else }
  End; { T_ABC_Context_Model.Set_Current_Frequencies }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Context_Model.Reset;
{ Resets all data                                                                     }
{-------------------------------------------------------------------------------------}
  Var
    I                   : Integer;
    Order               : Integer;
    Cumulated_Frequency : Integer;

  Begin { T_ABC_Context_Model.Reset }
    { Initialize lookup tables }
    For I := 0 To M_Number_Of_Symbols - 1 do
      Begin { For }
        M_A_Symbol_To_Index [I]     := I + 1;
        M_A_Index_To_Symbol [I + 1] := I;
      End; { For }

    { Frequency of 0 must be 0 }
    M_A_Frequency [0]               := 0;
    M_P_A_Context_Frequency_0^ [0]  := 0;
    For Order := 0 to M_Context_1_Mask do
      Begin { For }
        M_P_A_Context_Frequency_1^ [Order, 0] := 0;
      End; { For }
    For Order := 0 to M_Context_n_Mask do
      Begin { For }
        M_P_A_Context_Frequency_n^ [Order, 0] := 0;
      End; { For }

    { Initialize frequency array of order 0 }
    If ABC_MCP.Compress_Information.Fast_F = TRUE then
      Begin { then }
        For I := 1 to M_Number_Of_Symbols do
          Begin { For }
            M_P_A_Context_Frequency_0^ [I] := (M_Number_Of_Symbols + 1 - I) * M_Increment_0 * ABC_MODEL_INIT_FACTOR;
          End; { For }
      End { then }
        else
      Begin { else }
        For I := 1 to M_Number_Of_Symbols do
          Begin { For }
            M_P_A_Context_Frequency_0^ [I] := M_Increment_0 * ABC_MODEL_INIT_FACTOR;
          End; { For }
      End; { else }

    { Initialize frequency array of order 1 }
    If ABC_MCP.Compress_Information.Fast_F = TRUE then
      Begin { then }
        For Order := 0 to M_Context_1_Mask do
          Begin { For }
            For I := 1 to M_Number_Of_Symbols do
              Begin { For }
                M_P_A_Context_Frequency_1^ [Order, I]  := (M_Number_Of_Symbols + 1 - I) * M_Increment_1 * ABC_MODEL_INIT_FACTOR;
              End; { For }
          End; { For }
      End { then }
        else
      Begin { else }
        For Order := 0 to M_Context_1_Mask do
          Begin { For }
            For I := 1 to M_Number_Of_Symbols do
              Begin { For }
                M_P_A_Context_Frequency_1^ [Order, I]  := M_Increment_1 * ABC_MODEL_INIT_FACTOR;
              End; { For }
          End; { For }
      End; { else }

    { Initialize frequency array of order n }
    If ABC_MCP.Compress_Information.Fast_F = TRUE then
      Begin { then }
        For Order := 0 to M_Context_n_Mask do
          Begin { For }
            For I := 1 to M_Number_Of_Symbols do
              Begin { For }
                M_P_A_Context_Frequency_n^ [Order, I]  := (M_Number_Of_Symbols + 1 - I) * M_Increment_n * ABC_MODEL_INIT_FACTOR;
              End; { For }
          End; { For }
      End { then }
        else
      Begin { else }
        For Order := 0 to M_Context_n_Mask do
          Begin { For }
            For I := 1 to M_Number_Of_Symbols do
              Begin { For }
                M_P_A_Context_Frequency_n^ [Order, I]  := M_Increment_n * ABC_MODEL_INIT_FACTOR;
              End; { For }
          End; { For }
      End; { else }

    { Initialize cumulated frequency array of order 0 }
    Cumulated_Frequency := 0;
    For I := M_Number_Of_Symbols downto 0 do
      Begin { For }
        M_P_A_Context_Cumulated_Frequency_0^ [I] := Cumulated_Frequency;
        Inc (Cumulated_Frequency, M_P_A_Context_Frequency_0^ [I]);
      End; { For }

    { Initialize cumulated frequency array of order 1 }
    For Order := 0 to M_Context_1_Mask do
      Begin { For }
        Cumulated_Frequency := 0;
        For I := M_Number_Of_Symbols downto 0 do
          Begin { For }
            M_P_A_Context_Cumulated_Frequency_1^ [Order, I] := Cumulated_Frequency;
            Inc (Cumulated_Frequency, M_P_A_Context_Frequency_1^ [Order, I]);
          End; { For }
      End; { For }

    { Initialize cumulated frequency array of order n }
    For Order := 0 to M_Context_n_Mask do
      Begin { For }
        Cumulated_Frequency := 0;
        For I := M_Number_Of_Symbols downto 0 do
          Begin { For }
            M_P_A_Context_Cumulated_Frequency_n^ [Order, I] := Cumulated_Frequency;
            Inc (Cumulated_Frequency, M_P_A_Context_Frequency_n^ [Order, I]);
          End; { For }
      End; { For }

    { Initialize old data and context }
    M_Cache     := 0;
    M_Context_1 := 0;
    M_Context_n := 0;

    { Set current frequency array }
    Set_Current_Frequencies (0);

    { Initialize counter and old index }
    M_Increment_C := 0;
    M_Old_Index   := 0;
  End; { T_ABC_Context_Model.Reset }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Context_Model.Update (Index : Integer);
{ Updates the model                                                                   }
{-------------------------------------------------------------------------------------}
  Var
    I                   : Integer;
    Cumulated_Frequency : Integer;

  Begin { T_ABC_Context_Model.Update }
    { Update context 0 }

    { Increment frequency }
    Inc (M_P_A_Context_Frequency_0^ [Index], M_Increment_0);

    { Update frequencies on top }
    I := Index;
    While I > 0 do
      Begin { While }
        Dec (I);
        Inc (M_P_A_Context_Cumulated_Frequency_0^ [I], M_Increment_0);
      End; { While }

    { Rescale if cumulated frequency passes ABC_MODEL_FREQUENCY_MAX }
    If M_P_A_Context_Cumulated_Frequency_0^ [0] >= ABC_MODEL_FREQUENCY_MAX then
      Begin { then }
        Cumulated_Frequency := 0;
        For I := M_Number_Of_Symbols downto 0 do
          Begin { For }
            M_P_A_Context_Frequency_0^ [I]           := (M_P_A_Context_Frequency_0^ [I] + 1) shr 1;
            M_P_A_Context_Cumulated_Frequency_0^ [I] := Cumulated_Frequency;
            Inc (Cumulated_Frequency, M_P_A_Context_Frequency_0^ [I]);
          End; { For }
      End; { then }


    { Update context 1 }

    { Increment frequency }
    Inc (M_P_A_Context_Frequency_1^ [M_Context_1, Index], M_Increment_1);

    { Update frequencies on top }
    I := Index;
    While I > 0 do
      Begin { While }
        Dec (I);
        Inc (M_P_A_Context_Cumulated_Frequency_1^ [M_Context_1, I], M_Increment_1);
      End; { While }

    { Rescale if cumulated frequency passes ABC_MODEL_FREQUENCY_MAX }
    If M_P_A_Context_Cumulated_Frequency_1^ [M_Context_1, 0] >= ABC_MODEL_FREQUENCY_MAX then
      Begin { then }
        Cumulated_Frequency := 0;
        For I := M_Number_Of_Symbols downto 0 do
          Begin { For }
            M_P_A_Context_Frequency_1^ [M_Context_1, I]           := (M_P_A_Context_Frequency_1^ [M_Context_1, I] + 1) shr 1;
            M_P_A_Context_Cumulated_Frequency_1^ [M_Context_1, I] := Cumulated_Frequency;
            Inc (Cumulated_Frequency, M_P_A_Context_Frequency_1^ [M_Context_1, I]);
          End; { For }
      End; { then }


    { Update context n }

    { Increment frequency }
    Inc (M_P_A_Context_Frequency_n^ [M_Context_n, Index], M_Increment_n);

    { Update frequencies on top }
    I := Index;
    While I > 0 do
      Begin { While }
        Dec (I);
        Inc (M_P_A_Context_Cumulated_Frequency_n^ [M_Context_n, I], M_Increment_n);
      End; { While }

    { Rescale if cumulated frequency passes ABC_MODEL_FREQUENCY_MAX }
    If M_P_A_Context_Cumulated_Frequency_n^ [M_Context_n, 0] >= ABC_MODEL_FREQUENCY_MAX then
      Begin { then }
        Cumulated_Frequency := 0;
        For I := M_Number_Of_Symbols downto 0 do
          Begin { For }
            M_P_A_Context_Frequency_n^ [M_Context_n, I]           := (M_P_A_Context_Frequency_n^ [M_Context_n, I] + 1) shr 1;
            M_P_A_Context_Cumulated_Frequency_n^ [M_Context_n, I] := Cumulated_Frequency;
            Inc (Cumulated_Frequency, M_P_A_Context_Frequency_n^ [M_Context_n, I]);
          End; { For }
      End; { then }


    { Update context }
    Update_Context (Index);


    { Set current frequency array and cumulated frequency array }
    Inc (M_Increment_C);
    If (M_Increment_C >= ABC_MODEL_RESCALE_COUNT) or (Index <> M_Old_Index) then
      Begin { then }
        M_Increment_C := 0;
        Set_Current_Frequencies (Index);
      End { then }
        else
      Begin { else }
        { Update frequencies on top }
        I := Index;
        While I > 0 do
          Begin { While }
            Dec (I);
            Inc (M_A_Cumulated_Frequency [I], M_Increment_0 shr 1);
          End; { While }
      End; { else }
    M_Old_Index := Index;
  End; { T_ABC_Context_Model.Update }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Context_Model.Update_Context (Index : Integer);
{ Updates context                                                                     }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Context_Model.Update_Context }
    M_Cache     := ((M_Cache shl M_Context_Bit_Shift) or Longword (Index - 1));
    M_Context_1 := M_Cache and M_Context_1_Mask;
    M_Context_n := M_Cache and M_Context_n_Mask;
  End; { T_ABC_Context_Model.Update_Context }


{#####################################################################################}
{ T_ABC_Context_Level_0_Model                                                      }
{#####################################################################################}


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Context_Level_0_Model.Update_Context (Index : Integer);
{ Updates context                                                                     }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Context_Level_0_Model.Update_Context }
    M_Cache     := ((M_Cache shl M_Context_Bit_Shift) or Longword (Index - 1));
    If Index - 1 <= 1 then
      Begin { then }
        M_Cache := M_Cache and $FFFFFFFC;
      End; { then }
    M_Context_1 := M_Cache and M_Context_1_Mask;
    M_Context_n := M_Cache and M_Context_n_Mask;
  End; { T_ABC_Context_Level_0_Model.Update_Context }


{#####################################################################################}
{ T_ABC_Symbol_Model                                                               }
{#####################################################################################}


{-------------------------------------------------------------------------------------}
Constructor T_ABC_Symbol_Model.Create (Alphabet_Size : Integer);
{ Creates hierarchical model                                                          }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;
    J : Integer;

  Begin { T_ABC_Symbol_Model.Create }
    M_Alphabet_Size := Alphabet_Size;

    { Create model of 1. level }
    M_First_Level_Model := T_ABC_Model.Create (ABC_MODEL_FIRST_LEVEL_SYMBOL_N, ABC_MODEL_FIRST_LEVEL_INCREMENT_0 div 2);

    { Create model of 2. level }
    M_Third_Level_Model_N := ABC_MODEL_SECOND_LEVEL_SYMBOL_N;
    For I := 0 to ABC_MODEL_SECOND_LEVEL_SYMBOL_N - 1 do
      Begin { For }
        If ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I + 1] >= M_Alphabet_Size then
          Begin { then }
            M_Third_Level_Model_N := I + 1;
            Break;
          End; { then }
      End; { For }
    M_Second_Level_Model := T_ABC_Model.Create (M_Third_Level_Model_N, ABC_MODEL_SECOND_LEVEL_INCREMENT_0 div 3);

    { Create models of 3. level }
    For I := 0 to 2 - 1 do
      Begin { For }
        M_A_Third_Level_Model [I] := T_ABC_Model.Create (ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I + 1] - ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I],
        ABC_MODEL_A_THIRD_LEVEL_INCREMENT_0 [I] div 2);
      End; { For }
    For I := 2 to M_Third_Level_Model_N - 2 do
      Begin { For }
        M_A_Third_Level_Model [I] := T_ABC_Model.Create (ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I + 1] - ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I], ABC_MODEL_A_THIRD_LEVEL_INCREMENT_0 [I]);
      End; { For }
    M_A_Third_Level_Model [M_Third_Level_Model_N - 1] := T_ABC_Model.Create (M_Alphabet_Size - ABC_MODEL_A_THIRD_LEVEL_GROUP_START [M_Third_Level_Model_N - 1], ABC_MODEL_A_THIRD_LEVEL_INCREMENT_0 [M_Third_Level_Model_N - 1]);

    { Initialize lookup table of 2. level }
    For I := 0 to ABC_MODEL_SECOND_LEVEL_SYMBOL_N - 1 do
      Begin { For }
        For J := ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I] to ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I + 1] - 1 do
          Begin { For }
            M_A_Symbol_To_Second_Level_Group_Number [J] := I;
          End; { For }
      End; { For }
  End; { T_ABC_Symbol_Model.Create }


{-------------------------------------------------------------------------------------}
Destructor T_ABC_Symbol_Model.Destroy;
{ Frees hierarchical model                                                            }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;

  Begin { T_ABC_Symbol_Model.Destroy }
    { Frees all models }
    M_First_Level_Model.Free;
    M_Second_Level_Model.Free;
    For I := 0 to M_Third_Level_Model_N - 1 do
      Begin { For }
        M_A_Third_Level_Model [I].Free;
      End; { For }

    Inherited Destroy;
  End; { T_ABC_Symbol_Model.Destroy }


{-------------------------------------------------------------------------------------}
Function T_ABC_Symbol_Model.Get_Third_Level_Model (I : Integer) : T_ABC_Model;
{ Returns 3. level model                                                              }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Symbol_Model.Get_Third_Level_Model }
    Result := M_A_Third_Level_Model [I];
  End; { T_ABC_Symbol_Model.Get_Third_Level_Model }


{-------------------------------------------------------------------------------------}
Function T_ABC_Symbol_Model.Third_Level_Group_Offset_To_Symbol (Group_Number : Integer; Group_Index : Integer) : Integer;
{ Translates group number and offset to symbol                                        }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Symbol_Model.Third_Level_Group_Offset_To_Symbol }
    Result := ABC_MODEL_A_THIRD_LEVEL_GROUP_START [Group_Number] + Group_Index;
  End; { T_ABC_Symbol_Model.Third_Level_Group_Offset_To_Symbol }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Symbol_Model.Reset;
{ Resets all models                                                                   }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;

  Begin { T_ABC_Symbol_Model.Reset }
    { Reset 1. level }
    M_First_Level_Model.Reset;

    { Reset 2. level }
    M_Second_Level_Model.Reset;

    { Reset 3. level }
    For I := 0 to M_Third_Level_Model_N - 1 do
      Begin { For }
        M_A_Third_Level_Model [I].Reset;
      End; { For }
  End; { T_ABC_Symbol_Model.Reset }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Symbol_Model.Get_Hierarchical_Information (Symbol : Integer; Var Model_0 : T_ABC_Model; Var Symbol_0 : Integer; Var Model_1 : T_ABC_Model; Var Symbol_1 : Integer; Var Model_2 : T_ABC_Model; Var Symbol_2 : Integer);
{ Returns models and indices of level 1, 2 and 3                                      }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Symbol_Model.Get_Hierarchical_Information }
    If Symbol < ABC_MODEL_FIRST_LEVEL_ESCAPE_SYMBOL then
      Begin { then }
        Model_0   := M_First_Level_Model;
        Symbol_0  := Symbol;
        Model_1   := NIL;
        Symbol_1  := 0;
        Model_2   := NIL;
        Symbol_2  := 0;
      End { then }
        else
      Begin { else }
        Model_0   := M_First_Level_Model;
        Symbol_0  := ABC_MODEL_FIRST_LEVEL_ESCAPE_SYMBOL;
        Model_1   := M_Second_Level_Model;
        Symbol_1  := M_A_Symbol_To_Second_Level_Group_Number [Symbol];
        Model_2   := M_A_Third_Level_Model [Symbol_1];
        If Model_2.Number_Of_Symbols = 1 then
          Begin { then }
            Model_2   := NIL;
            Symbol_2  := 0;
          End { then }
            else
          Begin { else }
            Symbol_2 := Symbol - ABC_MODEL_A_THIRD_LEVEL_GROUP_START [Symbol_1];
          End; { else }
      End; { else }
  End; { T_ABC_Symbol_Model.Get_Hierarchical_Information }
 

{#####################################################################################}
{ T_ABC_Symbol_Context_Model                                                          }
{#####################################################################################}
 

{-------------------------------------------------------------------------------------}
Constructor T_ABC_Symbol_Context_Model.Create (Alphabet_Size : Integer);
{ Creates hirarchical model                                                           }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;
    J : Integer;

  Begin { T_ABC_Symbol_Context_Model.Create }
    M_Alphabet_Size := Alphabet_Size;

    { Create 1. level model }
    M_First_Level_Model := T_ABC_Context_Level_0_Model.Create (ABC_MODEL_FIRST_LEVEL_SYMBOL_N, ABC_MODEL_FIRST_LEVEL_INCREMENT_0, ABC_MODEL_FIRST_LEVEL_INCREMENT_1, ABC_MODEL_FIRST_LEVEL_INCREMENT_n, ABC_MODEL_FIRST_LEVEL_CONTEXT_BIT_SHIFT, ABC_MODEL_FIRST_LEVEL_n);

    { Create 2. level models }
    M_Third_Level_Model_N := ABC_MODEL_SECOND_LEVEL_SYMBOL_N;
    For I := 0 to ABC_MODEL_SECOND_LEVEL_SYMBOL_N - 1 do
      Begin { For }
        If ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I + 1] >= M_Alphabet_Size then
          Begin { then }
            M_Third_Level_Model_N := I + 1;
            Break;
          End; { then }
      End; { For }
    M_Second_Level_Model := T_ABC_Context_Model.Create (M_Third_Level_Model_N, ABC_MODEL_SECOND_LEVEL_INCREMENT_0, ABC_MODEL_SECOND_LEVEL_INCREMENT_1, ABC_MODEL_SECOND_LEVEL_INCREMENT_n, ABC_MODEL_SECOND_LEVEL_CONTEXT_BIT_SHIFT, ABC_MODEL_SECOND_LEVEL_n);

    { Create 3. level models }
    For I := 0 to 2 - 1 do
      Begin { For }
        M_A_Third_Level_Model [I] := T_ABC_Context_Model.Create (ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I + 1] - ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I],
        ABC_MODEL_A_THIRD_LEVEL_INCREMENT_0 [I], ABC_MODEL_A_THIRD_LEVEL_INCREMENT_0 [I] div 1, ABC_MODEL_A_THIRD_LEVEL_INCREMENT_0 [I] div 2,
        1, 5);
      End; { For }
    For I := 2 to M_Third_Level_Model_N - 2 do
      Begin { For }
        M_A_Third_Level_Model [I] := T_ABC_Model.Create (ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I + 1] - ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I], ABC_MODEL_A_THIRD_LEVEL_INCREMENT_0 [I]);
      End; { For }
    M_A_Third_Level_Model [M_Third_Level_Model_N - 1] := T_ABC_Model.Create (M_Alphabet_Size - ABC_MODEL_A_THIRD_LEVEL_GROUP_START [M_Third_Level_Model_N - 1], ABC_MODEL_A_THIRD_LEVEL_INCREMENT_0 [M_Third_Level_Model_N - 1]);

    { Initialize lookup table }
    For I := 0 to ABC_MODEL_SECOND_LEVEL_SYMBOL_N - 1 do
      Begin { For }
        For J := ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I] to ABC_MODEL_A_THIRD_LEVEL_GROUP_START [I + 1] - 1 do
          Begin { For }
            M_A_Symbol_To_Second_Level_Group_Number [J] := I;
          End; { For }
      End; { For }
  End; { T_ABC_Symbol_Context_Model.Create }


{#####################################################################################}
{ T_ABC_Exponent_Model                                                             }
{#####################################################################################}


{-------------------------------------------------------------------------------------}
Constructor T_ABC_Exponent_Model.Create;
{ Creates hirarchical model                                                           }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;
    J : Integer;

  Begin { T_ABC_Exponent_Model.Create }
    { Create 1. level model }
    M_First_Level_Model := T_ABC_Context_Model.Create (ABC_MODEL_EXPONENT_FIRST_LEVEL_SYMBOL_N, ABC_MODEL_EXPONENT_INCREMENT_0, ABC_MODEL_EXPONENT_INCREMENT_1, ABC_MODEL_EXPONENT_INCREMENT_n, ABC_MODEL_EXPONENT_CONTEXT_BIT_SHIFT, ABC_MODEL_EXPONENT_n);

    { Create 2. level models }
    For I := 0 to ABC_MODEL_EXPONENT_FIRST_LEVEL_SYMBOL_N - 2 do
      Begin { For }
        M_A_Second_Level_Model [I] := T_ABC_Model.Create (ABC_MODEL_A_SECOND_LEVEL_GROUP_START [I + 1] - ABC_MODEL_A_SECOND_LEVEL_GROUP_START [I], ABC_MODEL_A_SECOND_LEVEL_INCREMENT_0 [I]);
      End; { For }
    M_A_Second_Level_Model [ABC_MODEL_EXPONENT_FIRST_LEVEL_SYMBOL_N - 1] :=
      T_ABC_Context_Model.Create (ABC_MODEL_A_SECOND_LEVEL_GROUP_START [ABC_MODEL_EXPONENT_FIRST_LEVEL_SYMBOL_N - 1 + 1] - ABC_MODEL_A_SECOND_LEVEL_GROUP_START [ABC_MODEL_EXPONENT_FIRST_LEVEL_SYMBOL_N - 1],
      1300, 300, 80, 5, 2);

    { Initialize lookup table }
    For I := 0 to ABC_MODEL_EXPONENT_FIRST_LEVEL_SYMBOL_N - 1 do
      Begin { For }
        For J := ABC_MODEL_A_SECOND_LEVEL_GROUP_START [I] to ABC_MODEL_A_SECOND_LEVEL_GROUP_START [I + 1] - 1 do
          Begin { For }
            M_A_Symbol_To_Second_Level_Group_Number [J] := I;
          End; { For }
      End; { For }
  End; { T_ABC_Exponent_Model.Create }


{-------------------------------------------------------------------------------------}
Destructor T_ABC_Exponent_Model.Destroy;
{ Frees models                                                                        }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;

  Begin { T_ABC_Exponent_Model.Destroy }
    M_First_Level_Model.Free;
    For I := 0 to ABC_MODEL_EXPONENT_FIRST_LEVEL_SYMBOL_N - 1 do
      Begin { For }
        M_A_Second_Level_Model [I].Free;
      End; { For }

    Inherited Destroy;
  End; { T_ABC_Exponent_Model.Destroy }


{-------------------------------------------------------------------------------------}
Function T_ABC_Exponent_Model.Get_Second_Level_Model (I : Integer) : T_ABC_Model;
{ Returns 2. level model                                                              }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Exponent_Model.Get_Second_Level_Model }
    Result := M_A_Second_Level_Model [I];
  End; { T_ABC_Exponent_Model.Get_Second_Level_Model }


{-------------------------------------------------------------------------------------}
Function T_ABC_Exponent_Model.Second_Level_Group_Offset_To_Symbol (Group_Number : Integer; Group_Index : Integer) : Integer;
{ Returns associated symbol from group and offset                                     }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Exponent_Model.Second_Level_Group_Offset_To_Symbol }
    Result := ABC_MODEL_A_SECOND_LEVEL_GROUP_START [Group_Number] + Group_Index;
  End; { T_ABC_Exponent_Model.Second_Level_Group_Offset_To_Symbol }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Exponent_Model.Reset;
{ Resets all models                                                                   }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;

  Begin { T_ABC_Exponent_Model.Reset }
    { Reset 1. level }
    M_First_Level_Model.Reset;

    { Reset 2. level }
    For I := 0 to ABC_MODEL_EXPONENT_FIRST_LEVEL_SYMBOL_N - 1 do
      Begin { For }
        M_A_Second_Level_Model [I].Reset;
      End; { For }
  End; { T_ABC_Exponent_Model.Reset }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Exponent_Model.Get_Hierarchical_Information (Symbol : Integer; Var Model_0 : T_ABC_Model; Var Symbol_0 : Integer; Var Model_1 : T_ABC_Model; Var Symbol_1 : Integer);
{ Calculates models and offsets                                                       }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Exponent_Model.Get_Hierarchical_Information }
    Model_0   := M_First_Level_Model;
    Symbol_0  := M_A_Symbol_To_Second_Level_Group_Number [Symbol];
    Model_1   := M_A_Second_Level_Model [Symbol_0];
    If Model_1.Number_Of_Symbols = 1 then
      Begin { then }
        Model_1   := NIL;
        Symbol_1  := 0;
      End { then }
        else
      Begin { else }
        Symbol_1 := Symbol - ABC_MODEL_A_SECOND_LEVEL_GROUP_START [Symbol_0];
      End; { else }
  End; { T_ABC_Exponent_Model.Get_Hierarchical_Information }


{#####################################################################################}
{ T_ABC_Mantissa_Model                                                             }
{#####################################################################################}


{-------------------------------------------------------------------------------------}
Constructor T_ABC_Mantissa_Model.Create;
{ Creates a new model for mantissa data of upper bits                                 }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Mantissa_Model.Create }
    M_Model := T_ABC_Model.Create (ABC_BIT_N, 1);
  End; { T_ABC_Mantissa_Model.Create }


{-------------------------------------------------------------------------------------}
Destructor T_ABC_Mantissa_Model.Destroy;
{ Frees model                                                                         }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Mantissa_Model.Destroy }
    M_Model.Free;

    Inherited Destroy;
  End; { T_ABC_Mantissa_Model.Destroy }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_Mantissa_Model.Reset;
{ Resets model                                                                        }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Mantissa_Model.Reset }
    M_Model.Reset;
  End; { T_ABC_Mantissa_Model.Reset }


{#####################################################################################}
{ T_ABC_Mantissa_Model_1                                                           }
{#####################################################################################}


{-------------------------------------------------------------------------------------}
Constructor T_ABC_Mantissa_Model_1.Create;
{ Creates a new model for mantissa data of first bits                                 }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Mantissa_Model_1.Create }
    M_Model := T_ABC_Model.Create (2, 100);
  End; { T_ABC_Mantissa_Model_1.Create }


{#####################################################################################}
{ T_ABC_Mantissa_Model_2                                                           }
{#####################################################################################}


{-------------------------------------------------------------------------------------}
Constructor T_ABC_Mantissa_Model_2.Create;
{ Creates a new model for mantissa data of second bits                                }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Mantissa_Model_2.Create }
    M_Model := T_ABC_Model.Create (4, 40);
  End; { T_ABC_Mantissa_Model_2.Create }


{#####################################################################################}
{ T_ABC_Mantissa_Model_3                                                           }
{#####################################################################################}


{-------------------------------------------------------------------------------------}
Constructor T_ABC_Mantissa_Model_3.Create;
{ Creates a new model for mantissa data of third bits                                 }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_Mantissa_Model_3.Create }
    M_Model := T_ABC_Context_Model.Create (8, 40, 20, 10, 3, 3);
  End; { T_ABC_Mantissa_Model_3.Create }


End.


