{#####################################################################################}
{##                                                                                 ##}
{## ABC - Advanced Blocksorting Compressor                                          ##}
{##                                                                                 ##}
{## ABC_RLE_EXP_Unit                                                                ##}
{##                                                                                 ##}
{## Run-Length-Encoding with a logarithmic threshold run                            ##}
{##                                                                                 ##}
{## Copyright (C) 2002-2003 Jrgen Abel                                             ##}
{##                                                                                 ##}
{#####################################################################################}


{$I ABC_Include.pas}


Unit ABC_RLE_EXP_Unit;


Interface


Uses
  Classes, ABC_Global_Unit;


Type
  T_ABC_RLE_EXP = Class
  Protected
    M_Threshold : Integer;
  Public
    Procedure Encode (Var Compress_Information : T_R_ABC_Compress_Information);
    Procedure Decode (Var Compress_Information : T_R_ABC_Compress_Information);
    Property  Threshold : Integer Read M_Threshold Write M_Threshold;
  End; { T_ABC_RLE_EXP }


Implementation


Uses
  ABC_MCP_Unit;


{-------------------------------------------------------------------------------------}
Procedure T_ABC_RLE_EXP.Encode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Encodes a block                                                                     }
{-------------------------------------------------------------------------------------}
  Var
    Input_Index     : Integer;
    Output_Index    : Integer;
    Current_Symbol  : Byte;
    Run_Symbol      : Byte;
    Run_Start_Index : Integer;
    Run_Length      : Integer;
    N               : Integer;
    Exponent        : Integer;
    I               : Integer;

  Begin { T_ABC_RLE_EXP.Encode }
    If Compress_Information.P_Input_Record^.Size <= 0 then
      Begin { then }
        { Switch buffer }
        ABC_MCP.Switch_Block (Compress_Information);
        Exit;
      End; { then }

    { Copy first character }
    Input_Index                                                     := 0;
    Output_Index                                                    := 0;
    Run_Symbol                                                      := Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index];
    Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Run_Symbol;
    Run_Start_Index                                                 := Input_Index;
    Inc (Output_Index);
    Inc (Input_Index);

    { Process remaining symbols }
    While Input_Index <= Compress_Information.P_Input_Record^.Size - 1 do
      Begin { While }
        Current_Symbol := Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index];
        If Current_Symbol = Run_Symbol then
          Begin { then }
            Run_Length := (Input_Index + 1) - Run_Start_Index;
            If Run_Length < M_Threshold then
              Begin { then }
                { Output symbol }
                Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index] := Run_Symbol;
                Inc (Output_Index);
              End; { then }
          End { then }
            else
          Begin { else }
            Run_Length := Input_Index - Run_Start_Index;
            If Run_Length >= M_Threshold then
              Begin { then }
                N := Run_Length - M_Threshold;

                { Output exponent }
                Exponent := ABC_MCP.Get_Bin_Length (N);
                For I := 0 to Exponent do
                  Begin { For }
                    Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Run_Symbol;
                    Inc (Output_Index);
                  End; { For }

                { Output mantissa }
                If Exponent > 0 then
                  Begin { then }
                    Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.RLE_Mantissa_Index] := N;
                    Inc (Compress_Information.P_Binary_Block^.RLE_Mantissa_Index);
                  End; { then }
              End; { then }

            { Output symbol }
            Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Current_Symbol;
            Inc (Output_Index);
            Run_Symbol                                                      := Current_Symbol;
            Run_Start_Index                                                 := Input_Index;
          End; { else }
        Inc (Input_Index);
      End; { While }

    { Handle last run }
    Run_Length := Input_Index - Run_Start_Index;
    If Run_Length >= M_Threshold then
      Begin { then }
        N := Run_Length - M_Threshold;

        { Output exponent }
        Exponent := ABC_MCP.Get_Bin_Length (N);
        For I := 0 to Exponent do
          Begin { For }
            Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Run_Symbol;
            Inc (Output_Index);
          End; { For }

        { output mantissa }
        If Exponent > 0 then
          Begin { then }
            Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.RLE_Mantissa_Index] := N;
            Inc (Compress_Information.P_Binary_Block^.RLE_Mantissa_Index);
          End; { then }
      End; { then }

    { Set length }
    Compress_Information.P_Output_Record^.Size := Output_Index;
  End; { T_ABC_RLE_EXP.Encode }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_RLE_EXP.Decode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Decodes a block                                                                     }
{-------------------------------------------------------------------------------------}
  Var
    Input_Index   : Integer;
    Output_Index  : Integer;
    Symbol        : Integer;
    Run_Length    : Integer;
    I             : Integer;
    Exponent      : Integer;
    A_Temp_Index  : Array [0 .. ABC_MOST_SIGNIFICANT_BIT - 1] of Integer;

  Begin { T_ABC_RLE_EXP.Decode }
    If Compress_Information.P_Input_Record^.Size <= 0 then
      Begin { then }
        { Switch buffer }
        ABC_MCP.Switch_Block (Compress_Information);
        Exit;
      End; { then }

    { Reset all mantissa indices }
    Input_Index  := 0;
    While Input_Index <= Compress_Information.P_Input_Record^.Size - 1 do
      Begin { While }
        { Copy symbol }
        Symbol := Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index];
        Inc (Input_Index);

        { Check for run }
        If (Input_Index <= Compress_Information.P_Input_Record^.Size - 2) and (Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index] = Symbol) then
          Begin { then }
            Inc (Input_Index);

            { Calculate exponent }
            Run_Length := 2;
            While (Input_Index <= Compress_Information.P_Input_Record^.Size - 1) and (Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index] = Symbol) do
              Begin { While }
                Inc (Input_Index);
                Inc (Run_Length);
              End; { While }

            If Run_Length > M_Threshold then
              Begin { then }
                { Get run length }
                Exponent := Run_Length - M_Threshold;
                Dec (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
              End; { then }
          End; { then }
      End; { While }
    Move (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [0], A_Temp_Index [0], SizeOf (A_Temp_Index));

    { Process all symbols }
    Input_Index  := 0;
    Output_Index := 0;
    While Input_Index <= Compress_Information.P_Input_Record^.Size - 1 do
      Begin { While }
        { Copy symbol }
        Symbol                                                          := Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index];
        Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Symbol;
        Inc (Output_Index);
        Inc (Input_Index);

        { Check for run }
        If (Input_Index <= Compress_Information.P_Input_Record^.Size - 2) and (Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index] = Symbol) then
          Begin { then }
            Inc (Input_Index);

            { Calculate exponent }
            Run_Length := 2;
            While (Input_Index <= Compress_Information.P_Input_Record^.Size - 1) and (Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index] = Symbol) do
              Begin { While }
                Inc (Input_Index);
                Inc (Run_Length);
              End; { While }

            If Run_Length > M_Threshold then
              Begin { then }
                { Get run length }
                Exponent := Run_Length - M_Threshold;

                { Get mantissa }
                If Exponent > 0 then
                  Begin { then }
                    Run_Length := Compress_Information.P_Binary_Block^.P_RLE_Mantissa_Buffer^ [Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]] + ($01 shl Exponent) - 1;
                    Inc (Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [Exponent]);
                  End { then }
                    else
                  Begin { else }
                    Run_Length := 0;
                  End; { else }

                For I := 0 to M_Threshold + Run_Length - 2 do
                  Begin { For }
                    Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Symbol;
                    Inc (Output_Index);
                  End; { For }
              End { then }
                else
              Begin { else }
                For I := 0 to Run_Length - 2 do
                  Begin { For }
                    Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Symbol;
                    Inc (Output_Index);
                  End; { For }
              End; { else }
          End; { then }
      End; { While }

    { Reset all mantissa indices }
    Move (A_Temp_Index [0], Compress_Information.P_Binary_Block^.A_RLE_Mantissa_Index [0], SizeOf (A_Temp_Index));

    { Set length }
    Compress_Information.P_Output_Record^.Size := Output_Index;
  End; { T_ABC_RLE_EXP.Decode }


End.
