{#####################################################################################}
{##                                                                                 ##}
{## ABC - Advanced Blocksorting Compressor                                          ##}
{##                                                                                 ##}
{## ABC_SIF_Unit                                                                    ##}
{##                                                                                 ##}
{## Sorted Inversion Frequency transformation (MTF replacement)                     ##}
{##                                                                                 ##}
{## Copyright (C) 2002-2003 Jrgen Abel                                             ##}
{##                                                                                 ##}
{#####################################################################################}


{$I ABC_Include.pas}


Unit ABC_SIF_Unit;


Interface


Uses
  ABC_Global_Unit;


Type
  T_ABC_SIF = Class
  Protected
    M_A_Frequency         : Array [$00 .. ABC_BYTE_MAXIMUM_SYMBOL] of Integer;
    M_A_Sorted_Frequency  : Array [$00 .. ABC_BYTE_MAXIMUM_SYMBOL] of Integer;
    M_Alphabet_Size       : Integer;
    M_Alphabet_Max        : Integer;
    M_Sort_F              : Integer;
    M_Buffer_Size         : Integer;
    M_A_Last_Position     : Array [$00 .. ABC_BYTE_MAXIMUM_SYMBOL] of Integer;
    M_A_Index             : Array [$00 .. ABC_BYTE_MAXIMUM_SYMBOL] of Integer;
    M_P_IF_Buffer         : T_P_ABC_Integer_Array;
    M_A_AT                : Array [$00 .. ABC_BYTE_MAXIMUM_SYMBOL] of Byte;
    M_A_AIT               : Array [$00 .. ABC_BYTE_MAXIMUM_SYMBOL] of Byte;
    Procedure   Calculate_Symbol_Distribution (Var Compress_Information : T_R_ABC_Compress_Information);
  Public
    Procedure   Encode (Var Compress_Information : T_R_ABC_Compress_Information);
    Procedure   Decode (Var Compress_Information : T_R_ABC_Compress_Information);
  End; { T_ABC_SIF }


Implementation


Uses
  ABC_MCP_Unit;


Const
  ABC_SIF_THRESHOLD           = 10;   { Threshold for descending sorting }
  ABC_SIF_ASCENDING_FLAG      = 0;    { Flag for ascending sorting }
  ABC_SIF_DESCENDING_FLAG     = 1;    { Flag for descending sorting }
  ABC_SIF_NOT_USED            = $0;   { Constant for field not used }
  ABC_SIF_USED                = $1;   { Constant for field used }
  ABC_SIF_INDEX_START         = -1;   { Constant for start of index }
  ABC_SIF_UNDEFINED           = -1;   { Constant for undefinded value }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_SIF.Calculate_Symbol_Distribution (Var Compress_Information : T_R_ABC_Compress_Information);
{ Calculates alphabet size and share of frequent symbols                              }
{-------------------------------------------------------------------------------------}
  Var
    Symbol                  : Byte;
    Average_Frequency       : Integer;
    More_Frequent_Symbol_N  : Integer;

  Begin { T_ABC_SIF.Calculate_Symbol_Distribution }
    { Calculate alphabet size }
    M_Alphabet_Size := 0;
    For Symbol := 0 to M_Alphabet_Max - 1 do
      Begin { For }
        If M_A_Frequency [Symbol] > 0 then
          Begin { then }
            Inc (M_Alphabet_Size);
          End; { then }
      End; { For }

    { Calculate share of frequent symbols }
    Average_Frequency       := M_Buffer_Size div M_Alphabet_Size;
    More_Frequent_Symbol_N  := 0;
    For Symbol := 0 to M_Alphabet_Max - 1 do
      Begin { For }
        If M_A_Frequency [Symbol] > (Average_Frequency shl 1) then
          Begin { then }
            Inc (More_Frequent_Symbol_N);
          End; { then }
      End; { For }
    If More_Frequent_Symbol_N * 100 div M_Alphabet_Size >= ABC_SIF_THRESHOLD then
      Begin { then }
        M_Sort_F := ABC_SIF_ASCENDING_FLAG;
      End { then }
        else
      Begin { else }
        M_Sort_F := ABC_SIF_DESCENDING_FLAG;
      End; { else }
  End; { T_ABC_SIF.Calculate_Symbol_Distribution }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_SIF.Encode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Encodes a block                                                                     }
{-------------------------------------------------------------------------------------}
  Var
    I                 : Integer;
    Index             : Integer;
    Symbol            : Byte;
    {$IFNDEF SIF_NO_PERMUTATION}
    Last_Index        : Integer;
    Insert_Index      : Integer;
    {$ENDIF}
    {$IFDEF OPTIMIZATION_ON}
    Buffer_Size       : Integer;
    P_Input_Buffer    : Pointer;
    P_A_Last_Position : Pointer;
    P_A_Index         : Pointer;
    P_IF_Buffer       : Pointer;
    Alphabet_Max      : Integer;
    {$ELSE}
    Distance          : Integer;
    J                 : Integer;
    {$ENDIF}

  Begin { T_ABC_SIF.Encode }
    { Set pointer }
    M_P_IF_Buffer := Compress_Information.P_Temp_A_0;

    { Initialize arrays }
    FillChar (M_A_Frequency [0], SizeOf (M_A_Frequency), $00);
    FillChar (M_A_Last_Position [0], SizeOf (M_A_Last_Position), $FF);
    For I := 0 to ABC_BYTE_MAXIMUM_SYMBOL do
      Begin { For }
        M_A_AIT [I] := I;
      End; { For }

    { Calculate frequency distribution }
    For Index := 0 to Compress_Information.P_Input_Record^.Size - 1 do
      Begin { For }
        Symbol := Compress_Information.P_Input_Record^.P_Buffer^ [Index];
        Inc (M_A_Frequency [Symbol]);
      End; { For }

    { Calculate greatest alphabet symbol }
    M_Alphabet_Max := ABC_BYTE_MAXIMUM_SYMBOL;
    While (M_Alphabet_Max > 0) and (M_A_Frequency [M_Alphabet_Max] = 0) do
      Begin { While }
        Dec (M_Alphabet_Max);
      End; { While }
    Inc (M_Alphabet_Max);

    { Calculate alphabet size and share of frequent symbols }
    M_Buffer_Size := Compress_Information.P_Input_Record^.Size;
    Calculate_Symbol_Distribution (Compress_Information);

    {$IFNDEF SIF_NO_PERMUTATION}
    { Calculate permutation table }
    Index           := 0;
    Last_Index      := 0;
    For Symbol := 0 to M_Alphabet_Max - 1 do
      Begin { For }
        If M_A_Frequency [Symbol] > 0 then
          Begin { then }
            { Calculate insert position }
            Insert_Index := 0;
            If M_Sort_F = ABC_SIF_ASCENDING_FLAG then
              Begin { then }
                For I := Index - 1 downto 0 do
                  Begin { For }
                    If M_A_Frequency [Symbol] > M_A_Frequency [M_A_AIT [I]] then
                      Begin { then }
                        Insert_Index := I + 1;
                        Break;
                      End; { then }
                  End; { For }
              End { then }
                else
              Begin { else }
                For I := Index - 1 downto 0 do
                  Begin { For }
                    If M_A_Frequency [Symbol] < M_A_Frequency [M_A_AIT [I]] then
                      Begin { then }
                        Insert_Index := I + 1;
                        Break;
                      End; { then }
                  End; { For }
              End; { else }

            For I := Last_Index - 1 downto Insert_Index do
              Begin { For }
                M_A_AIT [I + 1] := M_A_AIT [I];
              End; { For }
            M_A_AIT [Insert_Index] := Symbol;

            Inc (Index);
            Inc (Last_Index);
          End { then }
            else
          Begin { else }
            { Add symbol }
            M_A_AIT [Last_Index] := Symbol;
            Inc (Last_Index);
          End; { else }
      End; { For }
    M_Alphabet_Size := Index;
    {$ELSE}
    M_Alphabet_Size := M_Alphabet_Max;
    {$ENDIF}

    { Output frequencies }
    ABC_MCP.Write_Binary_Data (M_Alphabet_Max);
    For I := 0 to M_Alphabet_Max - 1 do
      Begin { For }
        M_A_Sorted_Frequency [I] := M_A_Frequency [M_A_AIT [I]];
        ABC_MCP.Write_Binary_Data (M_A_Frequency [I]);
      End; { For }

    {$IFNDEF SIF_NO_PERMUTATION}
    { Initialize 1:1 transformation }
    For I := M_Alphabet_Max - 1 downto 0 do
      Begin { For }
        M_A_AT [M_A_AIT [I]] := I;
      End; { For }

    { Permutate all symbols }
    For I := 0 to Compress_Information.P_Input_Record^.Size - 1 do
      Begin { For }
        Compress_Information.P_Input_Record^.P_Buffer^ [I] := M_A_AT [Compress_Information.P_Input_Record^.P_Buffer^ [I]];
      End; { For }
    {$ENDIF}

    { Initialize indices }
    Index := 0;
    For I := 0 to M_Alphabet_Max - 1 do
      Begin { For }
        M_A_Index [I] := Index;
        Inc (Index, M_A_Sorted_Frequency [I]);
      End; { For }

    { Handle all symbols }
    {$IFDEF OPTIMIZATION_ON}
    Buffer_Size           := Compress_Information.P_Input_Record^.Size;
    P_Input_Buffer        := Compress_Information.P_Input_Record^.P_Buffer;
    P_A_Last_Position     := @M_A_Last_Position [0];
    P_A_Index             := @M_A_Index [0];
    P_IF_Buffer           := M_P_IF_Buffer;
    Alphabet_Max          := M_Alphabet_Max;
    Asm
      PUSH    EBX
      PUSH    ESI
      PUSH    EDI

      MOV     EDI,[P_A_Last_Position]
      XOR     ECX,ECX
@_Encode_00:
      MOV     ESI,[P_Input_Buffer]
      XOR     EAX,EAX
      MOV     AL,[ESI + ECX]
      MOV     EDX,ECX
      SUB     EDX,[EDI + EAX*4]
      DEC     EDX
      MOV     ESI,[P_A_Index]
      MOV     EBX,[ESI + EAX*4]
      INC     DWORD PTR [ESI + EAX*4]
      MOV     ESI,[P_IF_Buffer]
      MOV     [ESI + EBX*4],EDX

      MOV     [EDI + EAX*4],ECX

      MOV     EDX,[Alphabet_Max]
      MOV     EBX,EAX
      INC     EBX
      SUB     EDX,EBX
      JLE     @_Encode_02
      MOV     ESI,EDI
      ADD     ESI,EBX
      ADD     ESI,EBX
      ADD     ESI,EBX
      ADD     ESI,EBX
@_Encode_01:
      INC     DWORD PTR [ESI]
      ADD     ESI,4
      DEC     EDX
      JNZ     @_Encode_01

@_Encode_02:
      INC     ECX
      CMP     ECX,[Buffer_Size]
      JNZ     @_Encode_00

      POP     EDI
      POP     ESI
      POP     EBX
    End; { Asm }
    {$ELSE}
    For I := 0 to Compress_Information.P_Input_Record^.Size - 1 do
      Begin { For }
        { Get current symbol }
        Symbol := Compress_Information.P_Input_Record^.P_Buffer^ [I];

        { Get distance }
        Distance := I - M_A_Last_Position [Symbol] - 1;

        { Output difference }
        M_P_IF_Buffer^ [M_A_Index [Symbol]] := Distance;

        { Increment index }
        Inc (M_A_Index [Symbol]);

        { Update last positions }
        M_A_Last_Position [Symbol] := I;
        For J := Symbol + 1 to M_Alphabet_Max - 1 do
          Begin { For }
            Inc (M_A_Last_Position [J]);
          End; { For }
      End; { For }
    {$ENDIF}

    { Ouput distances }
    For I := 0 to Index - 1 - M_A_Sorted_Frequency [M_Alphabet_Size - 1] do
      Begin { For }
        ABC_MCP.Write_Binary_Data (M_P_IF_Buffer^ [I]);
      End; { For }

    { Set length }
    Compress_Information.P_Output_Record^.Size := 0;

    { Save block data }
    ABC_MCP.Push_Binary_Block;
End; { T_ABC_SIF.Encode }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_SIF.Decode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Decodes a block                                                                     }
{-------------------------------------------------------------------------------------}
  Var
    I                     : Integer;
    Index                 : Integer;
    Symbol                : Integer;
    {$IFNDEF SIF_NO_PERMUTATION}
    Last_Index            : Integer;
    Insert_Index          : Integer;
    {$ENDIF}
    {$IFDEF OPTIMIZATION_ON}
    Mantissa_Size         : Integer;
    P_Exponent_Buffer     : Pointer;
    Exponent_Index        : Integer;
    P_Mantissa_Buffer     : Pointer;
    P_A_Mantissa_Index    : Pointer;
    P_A_Last_Position     : Pointer;
    Alphabet_Size         : Integer;
    P_A_Index             : Pointer;
    P_IF_Buffer           : Pointer;
    P_Output_Buffer       : Pointer;
    P_A_AIT               : Pointer;
    P_A_Sorted_Frequency  : Pointer;
    Buffer_Size           : Integer;
    {$ENDIF}

  Begin { T_ABC_SIF.Decode }
    { Set pointer }
    M_P_IF_Buffer := Compress_Information.P_Temp_A_0;

    { Restore block data }
    ABC_MCP.Pop_Binary_Block;

    { Initialize array }
    For I := 0 to ABC_BYTE_MAXIMUM_SYMBOL do
      Begin { For }
        M_A_AIT [I] := I;
      End; { For }

    { Input frequencies of symbols }
    M_Alphabet_Max := ABC_MCP.Read_Binary_Data;
    For I := 0 to M_Alphabet_Max - 1 do
      Begin { For }
        M_A_Frequency [I] := ABC_MCP.Read_Binary_Data;
      End; { For }

    { Calcualte output size }
    Compress_Information.P_Output_Record^.Size := 0;
    For I := 0 to M_Alphabet_Max - 1 do
      Begin { For }
        Inc (Compress_Information.P_Output_Record^.Size, M_A_Frequency [I]);
      End; { For }

    {$IFNDEF SIF_NO_PERMUTATION}
    { Calculate alphabet size and share of frequent symbols }
    M_Buffer_Size := Compress_Information.P_Output_Record^.Size;
    Calculate_Symbol_Distribution (Compress_Information);

    { Calculate permutation table }
    Index           := 0;
    Last_Index      := 0;
    For Symbol := 0 to M_Alphabet_Max - 1 do
      Begin { For }
        If M_A_Frequency [Symbol] > 0 then
          Begin { then }
            { Calculate insert position }
            Insert_Index := 0;
            If M_Sort_F = ABC_SIF_ASCENDING_FLAG then
              Begin { then }
                For I := Index - 1 downto 0 do
                  Begin { For }
                    If M_A_Frequency [Symbol] > M_A_Frequency [M_A_AIT [I]] then
                      Begin { then }
                        Insert_Index := I + 1;
                        Break;
                      End; { then }
                  End; { For }
              End { then }
                else
              Begin { else }
                For I := Index - 1 downto 0 do
                  Begin { For }
                    If M_A_Frequency [Symbol] < M_A_Frequency [M_A_AIT [I]] then
                      Begin { then }
                        Insert_Index := I + 1;
                        Break;
                      End; { then }
                  End; { For }
              End; { else }

            For I := Last_Index - 1 downto Insert_Index do
              Begin { For }
                M_A_AIT [I + 1] := M_A_AIT [I];
              End; { For }
            M_A_AIT [Insert_Index] := Symbol;

            Inc (Index);
            Inc (Last_Index);
          End { then }
            else
          Begin { else }
            { Add symbol }
            M_A_AIT [Last_Index] := Symbol;
            Inc (Last_Index);
          End; { else }
      End; { For }
    M_Alphabet_Size := Index;
    {$ELSE}
    M_Alphabet_Size := M_Alphabet_Max;
    {$ENDIF}

    { Calculate sorted frequencies }
    For I := 0 to M_Alphabet_Max - 1 do
      Begin { For }
        M_A_Sorted_Frequency [I] := M_A_Frequency [M_A_AIT [I]];
      End; { For }

    {$IFDEF OPTIMIZATION_ON}
    Mantissa_Size         := Compress_Information.P_Output_Record^.Size - M_A_Sorted_Frequency [M_Alphabet_Size - 1];
    P_Exponent_Buffer     := Compress_Information.P_Binary_Block^.P_Exponent_Buffer;
    Exponent_Index        := Compress_Information.P_Binary_Block^.Exponent_Index;
    P_IF_Buffer           := M_P_IF_Buffer;
    P_Mantissa_Buffer     := Compress_Information.P_Binary_Block^.P_Mantissa_Buffer;
    P_A_Mantissa_Index    := @Compress_Information.P_Binary_Block^.A_Mantissa_Index [0];
    P_A_Index             := @M_A_Index [0];
    Alphabet_Size         := M_Alphabet_Size;
    P_A_Sorted_Frequency  := @M_A_Sorted_Frequency [0];
    P_A_Last_Position     := @M_A_Last_Position [0];
    P_Output_Buffer       := Compress_Information.P_Output_Record^.P_Buffer;
    P_A_AIT               := @M_A_AIT [0];
    Buffer_Size           := Compress_Information.P_Output_Record^.Size;
    Asm
      PUSH    EBX
      PUSH    ESI
      PUSH    EDI

      { Input all distances }
      MOV     ESI,[P_Exponent_Buffer]
      MOV     EDI,[P_IF_Buffer]
      MOV     EBX,[P_Mantissa_Buffer]
      MOV     EDX,[P_A_Mantissa_Index]
      MOV     ECX,[Mantissa_Size]
      JECXZ   @_Decode_10

@_Decode_00:
      MOV     EAX,[Exponent_Index]
      INC     [Exponent_Index]
      MOVZX   EAX,BYTE PTR [ESI + EAX]
      TEST    EAX,EAX
      JZ      @_Decode_01
      PUSH    ESI
      MOV     ESI,[EDX + EAX*4]
      INC     DWORD PTR [EDX + EAX*4]
      MOV     ESI,[EBX + ESI*4]
      PUSH    ECX
      MOV     CL,AL
      XOR     EAX,EAX
      INC     EAX
      SHL     EAX,CL
      DEC     EAX
      POP     ECX
      ADD     EAX,ESI
      MOV     [EDI],EAX
      POP     ESI
      JMP     @_Decode_02

@_Decode_01:
      MOV     DWORD PTR [EDI],0

@_Decode_02:
      ADD     EDI,4
      DEC     ECX
      JNZ     @_Decode_00


      { Set indices }
@_Decode_10:
      MOV     ESI,[P_A_Sorted_Frequency]
      MOV     EDI,[P_A_Index]
      XOR     EAX,EAX
      XOR     ECX,ECX

@_Decode_11:
      MOV     [EDI + ECX*4],EAX
      ADD     EAX,[ESI + ECX*4]
      INC     ECX
      CMP     ECX,[Alphabet_Size]
      JC      @_Decode_11


      { Set A_Last_Position }
      MOV     EBX,[P_A_Index]
      MOV     ESI,[P_IF_Buffer]
      MOV     EDI,[P_A_Last_Position]

      MOV     ECX,[Alphabet_Size]
      SUB     ECX,2
@_Decode_20:
      MOV     EAX,[EBX + ECX*4]
      MOV     EAX,[ESI + EAX*4]
      MOV     [EDI + ECX*4],EAX
      DEC     ECX
      JNS     @_Decode_20


      { Handle all symbols }
@_Decode_30:
      XOR     EDX,EDX
      DEC     [Alphabet_Size]

@_Decode_31:
      MOV     EBX,[Alphabet_Size]
      MOV     ECX,EBX
      MOV     ESI,[P_A_Last_Position]
      JECXZ   @_Decode_40

@_Decode_32:
      CMP     [ESI],EDX
      JZ      @_Decode_34
@_Decode_33:
      ADD     ESI,4
      DEC     ECX
      JNZ     @_Decode_32
      JMP     @_Decode_40

@_Decode_34:
      MOV     EDI,[P_A_Last_Position]
      SUB     EDI,ESI
      NEG     EDI
      ADD     EDI,[P_A_Sorted_Frequency]
      CMP     DWORD PTR [EDI],0
      JZ      @_Decode_33
      SUB     EBX,ECX


@_Decode_40:
      {$IFNDEF SIF_NO_PERMUTATION}
      MOV     ESI,[P_A_AIT]
      MOV     AL,[ESI + EBX]
      {$ELSE}
      MOV     AL,BL
      {$ENDIF}
      MOV     ESI,[P_Output_Buffer]
      MOV     [ESI + EDX],AL
      INC     EDX

      CMP     EBX,[Alphabet_Size]
      JNC     @_Decode_41
      MOV     ESI,[P_A_Index]
      MOV     EDI,[P_A_Sorted_Frequency]
      INC     DWORD PTR [ESI + EBX*4]
      DEC     DWORD PTR [EDI + EBX*4]
      MOV     EAX,[ESI + EBX*4]
      MOV     ESI,[P_IF_Buffer]
      MOV     EAX,[ESI + EAX*4]
      MOV     ESI,[P_A_Last_Position]
      ADD     EAX,EDX
      MOV     [ESI + EBX*4],EAX

@_Decode_41:
      INC     EBX
      MOV     ECX,[Alphabet_Size]
      SUB     ECX,EBX
      JLE     @_Decode_43
      MOV     ESI,[P_A_Last_Position]
      LEA     ESI,ESI + EBX*4

@_Decode_42:
      INC     DWORD PTR [ESI]
      ADD     ESI,4
      DEC     ECX
      JNZ     @_Decode_42

@_Decode_43:
      CMP     EDX,[Buffer_Size]
      JC      @_Decode_31


@_Decode_50:
      POP     EDI
      POP     ESI
      POP     EBX
    End; { Asm }
    Compress_Information.P_Binary_Block^.Exponent_Index := Exponent_Index;
    {$ELSE}
    { Input all distances }
    For I := 0 to Compress_Information.P_Output_Record^.Size - 1 - M_A_Sorted_Frequency [M_Alphabet_Size - 1] do
      Begin { For }
        M_P_IF_Buffer^ [I] := ABC_MCP.Read_Binary_Data;
      End; { For }

    { Set indices }
    Index := 0;
    For I := 0 to M_Alphabet_Size - 1 do
      Begin { For }
        M_A_Index [I] := Index;
        Inc (Index, M_A_Sorted_Frequency [I]);
      End; { For }

    { Handle all symbols }
    For I := 0 to M_Alphabet_Size - 2 do
      Begin { For }
        M_A_Last_Position [I] := M_P_IF_Buffer^ [M_A_Index [I]];
      End; { For }
    Index := 0;
    While Index <= Compress_Information.P_Output_Record^.Size - 1 do
      Begin { While }
        Symbol := M_Alphabet_Size - 1;
        For I := 0 to M_Alphabet_Size - 2 do
          Begin { For }
            If (M_A_Last_Position [I] = Index) and (M_A_Sorted_Frequency [I] > 0) then
              Begin { then }
                Symbol := I;
                Break;
              End; { then }
          End; { For }

        {$IFNDEF SIF_NO_PERMUTATION}
        Compress_Information.P_Output_Record^.P_Buffer^ [Index] := M_A_AIT [Symbol];
        {$ELSE}
        Compress_Information.P_Output_Record^.P_Buffer^ [Index] := Symbol;
        {$ENDIF}
        Inc (Index);
        If Symbol < M_Alphabet_Size - 1 then
          Begin { then }
            Inc (M_A_Index [Symbol]);
            Dec (M_A_Sorted_Frequency [Symbol]);
            M_A_Last_Position [Symbol] := M_P_IF_Buffer^ [M_A_Index [Symbol]] + Index;
          End; { then }

        For I := Symbol + 1 to M_Alphabet_Size - 2 do
          Begin { For }
            Inc (M_A_Last_Position [I]);
          End; { For }
      End; { While }
    {$ENDIF}
  End; { T_ABC_SIF.Decode }


End.

